port = 8000 
upload_dir = 'web'
hot_reload = True
appname = 'LLM test'
autotest = '*'
#lm = 'host', "qwen3.5:9b" ,"http://localhost:11434/v1"
#llm = 'groq', 'llama-3.1-70b-versatile'
llm = 'google', "gemini-2.5-flash"
#llm = 'host', "https://openrouter.ai/api/v1", "OPENROUTER1_API_KEY", "deepseek/deepseek-chat-v3-0324:free"
#llm = 'openai', "gpt-4o"
#llm = 'mistral', "mistral-large-latest"
#llm_cache = 'llm.cache'



from unisi import *
from data import *
name = 'Linked tables'
order = 1

utable = Table("Users", id = 'User', limit = 150, ids = True,
    rows= users, headers=['name', 'age', 'height'])
otable = Table("Orders", id = 'Orders', limit = 150, ids = True, 
    rows= orders, headers=['name', 'sum'], link = (utable, {'type' : 'string', 'weight' : 'double'}))
blocks = [Block('TBlock', [], [utable, otable])]


# Copyright © 2024 UNISI Tech. All rights reserved.
from aiohttp import web, WSMsgType
from .users import *
from pathlib import Path
from .reloader import active_reloader 
from .autotest import recorder, run_tests
from .common import  *
from .llmrag import setup_llmrag
from .dbunits import dbupdates
from .kdb import Database
from config import port, upload_dir
import traceback, json, random, string
from urllib.parse import parse_qs

def generate_random_string(length=10):
    characters = string.ascii_letters + string.digits 
    return ''.join(random.choices(characters, k=length))

def context_user():
    return context_object(User)

def context_screen():
    user = context_user()
    return user.screen if user else None

def message_logger(message, type = 'error'):
    user = context_user()
    if user:    
        user.log(message, type)
    else:
        with logging_lock:
            logging.error(message)

Unishare.context_user = context_user
Unishare.message_logger = message_logger
User.type = User    

if config.db_dir:
    Unishare.db = Database(config.db_dir, message_logger) 

def make_user(request):
    parsed_query = parse_qs(request.query_string)
    if 'session' in parsed_query:
        session = parsed_query['session'][0]
        user_id = session.split('-')[1]
    else:
        user_id = parsed_query.get('id', [User.count])[0]
        session = f'{generate_random_string()}-{user_id}'      
          
    if config.share and 'session' in parsed_query:
        user = Unishare.sessions.get(session, None)
        if not user:
            error = f'Session id "{session}" is unknown. Connection refused!'
            with logging_lock:
                logging.error(error)
            return None, Error(error)
        user = User.type(session, user)
        ok = user.screens
    elif config.mirror and User.count:
        user = User.type(session, User.last_user)
        ok = user.screens
    elif not User.count:
        user = User.last_user        
        user.session = session
        user.monitor(session)
        ok = True
    else:
        user = User.type(session)
        ok = user.load()                  
    User.count += 1
    Unishare.sessions[session] = user 
    return user, ok

def handle(unit, event):
    handler_map = User.last_user.handlers        
    def h(fn):
        key = unit, event        
        func = handler_map.get(key, None)        
        if func:
            handler_map[key] =  compose_handlers(func, fn)  
        else: 
            handler_map[key] = fn
        return fn
    return h

Unishare.handle = handle

Unishare.test_list = []

def test(fn):    
    Unishare.test_list.append(fn)
    return fn

async def post_handler(request):
    reader = await request.multipart()
    field = await reader.next()   
    filename = upload_path(field.filename)      
    size = 0
    with open(filename, 'wb') as f:
        while True:
            chunk = await field.read_chunk()  
            if not chunk:
                break
            size += len(chunk)
            f.write(chunk)
    return web.Response(text=filename)

async def static_serve(request):    
    rpath = request.path    
    file_path  = Path(f"{webpath}{rpath}" )
    if request.path == '/':
        file_path /= 'index.html'
    if not file_path.exists():
        file_path = None
        #unmask win path
        if rpath.startswith('/') and rpath[2] == ':':
            rpath = rpath[1:]        
        for dir in config.public_dirs:              
            if rpath.startswith(dir):                
                if os.path.exists(rpath):
                    file_path  = Path(rpath)
                break            
    return web.FileResponse(file_path) if file_path else web.HTTPNotFound()
     
async def websocket_handler(request):
    ws = web.WebSocketResponse()
    await ws.prepare(request)    
    user, status = make_user(request)
    if not user:
        await ws.send_str(toJson(status))
    else:
        async def send(res):
            if type(res) != str:
                res = toJson(user.prepare_result(res))        
            await ws.send_str(res)        
        user.send = send         
        await send(True if status else empty_app) 
        try:
            async for msg in ws:
                if msg.type == WSMsgType.TEXT:
                    if msg.data == 'close':
                        await ws.close()
                    else:
                        raw_message = json.loads(msg.data)
                        message = None
                        if isinstance(raw_message, list):
                            if raw_message:
                                for raw_submessage in raw_message:
                                    message = ReceivedMessage(raw_submessage)                    
                                    result = await user.result4message(message)
                            else:                                
                                result = Warning('Empty command batch!')
                        else:                    
                            message = ReceivedMessage(raw_message)            
                            result = await user.result4message(message)                    
                        await send(result)
                        if message:
                            if recorder.record_file:
                                recorder.accept(message, user.prepare_result (result))
                            await user.reflect(message, result)     
                        if dbupdates:
                            await user.sync_dbupdates()                       
                elif msg.type == WSMsgType.ERROR:
                    user.log('ws connection closed with exception %s' % ws.exception())
        except BaseException as e:
            if not isinstance(e, ConnectionResetError):
                user.log(traceback.format_exc())
        await user.delete()   
    return ws     

def ensure_directory_exists(directory_path):
    if not os.path.exists(directory_path):
        os.makedirs(directory_path)
        print(f"Directory '{directory_path}' created.")

def start(user_type = User, http_handlers = []):    
    ensure_directory_exists(screens_dir)
    ensure_directory_exists(blocks_dir)
    setup_llmrag()

    User.type = user_type        
    run_tests(User.init_user())
                        #http_handlers has to be the first argument
    server_handlers = http_handlers + [web.get('/ws', websocket_handler), 
            web.static(f'/{config.upload_dir}', upload_dir), 
        web.get('/{tail:.*}', static_serve), web.post('/', post_handler)] 

    app = web.Application()
    app.add_routes(server_handlers)    
    web.run_app(app, port = port)
    



# Copyright © 2024 UNISI Tech. All rights reserved.
import os, platform, requests, logging
from .common import set_defaults
from .containers import Screen

blocks_dir = 'blocks'        
screens_dir =  'screens'        
testdir = 'autotest'

divpath = '\\' if platform.system() == 'Windows' else '/'
libpath = os.path.dirname(os.path.realpath(__file__))
webpath = f'{libpath}{divpath}web' 
app_dir = os.getcwd()

try:
    import config
except:
    if os.path.exists('config.py'):
        print('Invalid script is started! It has to be in a working directory.')
        exit()
    f = open('config.py', 'w')  
    f.write("""port = 8000 
upload_dir = 'web'
hot_reload  = True
logfile  = 'log'
autotest = '*'
appname = 'Unisi app'
""")
    f.close()
    import config
    print("Config with default parameters is created!")

#setting config variables
set_defaults(config,  dict(
    autotest= False,
    appname = 'Unisi app',
    upload_dir = 'web',
    logfile= None,
    hot_reload = False,    
    mirror = False,
    share = False,
    profile = 0, 
    llm = None,
    froze_time= None,
    monitor_tick = 0.005,
    pool = None,
    db_dir = None,
    lang = 'en-US',
    public_dirs = [],
    debug = False,
    image = 'icons/favicon-32x32.png'
))

Screen.defaults = dict(
    icon = None,
    prepare = None,            
    blocks = [],
    header = config.appname,                        
    toolbar = [], 
    order = 0,
    reload = config.hot_reload, 
    lang = config.lang,
    voice = not config.mirror,
    image = config.image
)

if config.froze_time == 0:
    print('froze_time in config.py can not be 0!')
    config.froze_time = None

def filename2url(fn):   
    if fn[0] == '/' or fn[1] == ':': #if full path
        fn = fn[len(app_dir):]   
    if fn[0] == divpath:
        fn = fn[1:]
    return fn 

def url2filepath(url):
    return url[url.find('/') + 1:].replace('%20',' ')   

def url2filename(url):
    return url[url.rfind('/') + 1:].replace('%20',' ')   

def upload_path(fpath):
    return f'{config.upload_dir}{divpath}{fpath}'
    
def cache_url(url):
    """cache url file in upload_dir and returns the local file name"""
    fname = url2filename(url)   
    fname = upload_path(fname)
    response = requests.get(url)
    if response.status_code != 200:
        return None
    file = open(fname, "wb")
    file.write(response.content)
    file.close() 
    return fname

def start_logging(): 
    format = "%(asctime)s - %(levelname)s - %(message)s"
    logfile = config.logfile
    handlers = [logging.FileHandler(logfile), logging.StreamHandler()] if logfile else []
    logging.basicConfig(level = logging.WARNING, format = format, handlers = handlers)    

start_logging()






# Copyright © 2024 UNISI Tech. All rights reserved.
from .utils import *
from .units import *
from .common import *
from .voicecom import VoiceCom
from .containers import Dialog, Screen
from .multimon import notify_monitor, logging_lock, run_external_process
from .dbunits import dbshare, dbupdates
import sys, asyncio, logging, importlib

class User:          
    last_user = None
    toolbar = []    
    count = 0
    
    def __init__(self, session: str, share = None):          
        self.session = session        
        self.active_dialog = None        
        self.last_message = None                               
        self.changed_units = set()
        self.voice = None
        self.screen_module = None

        if share:            
            self.screens = share.screens       
            if share.screens:     
                self.screen_module = share.screens[0] if config.mirror else share.screen_module
            
            self.handlers =  share.handlers        
            
            if share.reflections:            
                share.reflections.append(self)
            else:
                share.reflections = [share, self]                    
            self.reflections = share.reflections
        else:
            self.screens = []        
            self.reflections = []            
            self.handlers = {} 

        User.last_user = self
        self.monitor(session, share)     

    @property
    def sorted_changed_units(self):
        """sort changed units by type, 'block' types have priority"""
        if len(self.changed_units) > 1:
            not_block = lambda u: u.type != 'block' 
            if any(not not_block(x) for x in self.changed_units):
                return sorted(self.changed_units, key=not_block)  
        return self.changed_units       
        
    async def run_process(self, long_running_task, *args, progress_callback = None, **kwargs):
        if progress_callback and notify_monitor and progress_callback != self.progress: #progress notifies the monitor
            async def new_callback(value):
                asyncio.gather(notify_monitor('e', self.session, self.last_message), progress_callback(value))                
            progress_callback = new_callback
        return await run_external_process(long_running_task, *args, progress_callback = progress_callback, **kwargs)

    async def broadcast(self, message):
        screen = self.screen_module
        if type(message) != str:
            message = toJson(self.prepare_result(message))  
        await asyncio.gather(*[user.send(message)
            for user in self.reflections
                if user is not self and screen is user.screen_module])        
        
    async def reflect(self, message, result):
        if self.reflections and not message.screen_type:                        
            if result:
                await self.broadcast(result)        
            if message:                    
                msg_object = self.find_element(message)                                     
                if not isinstance(result, Message) or not result.contains(msg_object):                                                        
                    await self.broadcast(msg_object)    

    async def progress(self, value, *updates):
        """open or update progress window if str != null else close it """  
        if not self.testing:           
            msg = TypeMessage('progress', str(value), *updates, user = self)            
            await asyncio.gather(self.send(msg), self.reflect(None, msg))
            if notify_monitor:
                await notify_monitor('e', self.session, self.last_message) 
                                           
    def load_screen(self, file):                     
        name = file[:-3]        
        path = f'{screens_dir}{divpath}{file}'                
        spec = importlib.util.spec_from_file_location(name,path)
        module = importlib.util.module_from_spec(spec)        
        module.user = self  
                
        spec.loader.exec_module(module)            
        screen = Screen(getattr(module, 'name', ''))        
        #set system vars
        for var, val in screen.defaults.items():                                            
            setattr(screen, var, getattr(module, var, val))         
        if not isinstance(screen.blocks, list):
            screen.blocks = [screen.blocks]
            
        if User.toolbar and User.toolbar[0] not in screen.toolbar:
            screen.toolbar += User.toolbar
        
        if User.count > 0:
            screen.set_reactivity(self)        
        module.screen = screen#ChangedProxy(screen, screen)                                 
        return module
    
    async def delete(self):
        uss = Unishare.sessions
        if uss and uss.get(self.session):
            del uss[self.session]        
        if self.reflections: #reflections is common array
            if len(self.reflections) == 2: 
                self.reflections.clear() #1 element in user.reflections has no sense
            else:
                self.reflections.remove(self)  
        if notify_monitor:
            await notify_monitor('-', self.session, self.last_message)   
        if config.share:
            self.log(f'User is disconnected, session: {self.session}', type = 'info')            

    def set_clean(self):
        """remove user modules from sys """
        if os.path.exists(blocks_dir):
            for file in os.listdir(blocks_dir):
                if file.endswith(".py") and file != '__init__.py':
                    name = f'{blocks_dir}.{file[0:-3]}'
                    if name in sys.modules:
                        sys.modules[name].user = self
                        del sys.modules[name]                          
    def load(self):              
        if os.path.exists(screens_dir):
            for file in os.listdir(screens_dir):
                if file.endswith(".py") and file != '__init__.py':
                    module = self.load_screen(file)                
                    self.screens.append(module)                
        
        if self.screens:                        
            self.screens.sort(key=lambda s: s.screen.order)            
            main = self.screens[0]
            self.screen_module = main
            if hasattr(main, 'prepare'):  
                main.prepare()            
            self.update_menu()
            self.set_clean()                               
            return True                 

    def update_menu(self):
        menu = [[getattr(s, 'name', ''),getattr(s,'icon', None)] for s in self.screens]        
        for s in self.screens:
            s.screen.menu = menu

    @property
    def testing(self):        
        return  self.session == testdir
    
    @property
    def id(self):        
        return  self.session.split('-')[-1]
    
    @property
    def screen(self):        
        return  self.screen_module.screen if self.screen_module else empty_app

    def set_screen(self,name): 
        return self.screen_process(ArgObject(block = 'root', element = None, value = name))
    
    async def result4message(self, message):
        result = None        
        self.last_message = message     
        if dialog := self.active_dialog:            
            if message.element is None: #dialog command button is pressed
                self.active_dialog = None    
                if self.reflections:            
                    await self.broadcast(close_message)                                    
                result = await self.eval_handler(dialog.changed, dialog, message.value)
            else:
                el = self.find_element(message)
                if el:
                    result = await self.process_element(el, message)                        
        else:
            result = await self.process(message)           
        if result and isinstance(result, Dialog):
            self.active_dialog = result
        return result

    async def eval_handler(self, handler, *params):
        if notify_monitor:
            await notify_monitor('+', self.session, self.last_message)        
        result = await call_anysync(handler, *params)
        if notify_monitor:
            await notify_monitor('-', self.session, None)        
        return result
    
    def register_changed_unit(self, unit, property = None, value = None):
        """add unit to changed_units if it is changed outside of message"""
        if property == 'value':
            property = 'changed'
        m = self.last_message
        
        if m and m.event == 'modify' and m.element == unit.name and (epath := 
                self.find_path(unit)) and m.block == strpath(epath):            
            return False
        if not m or m.element != unit.name or property != m.event or value != m.value:
            self.changed_units.add(unit)            

    @property
    def blocks(self):
        return [self.active_dialog, *self.screen.blocks] if self.active_dialog and \
            self.active_dialog.value else self.screen.blocks

    def find_element(self, message):                       
        elname = message.element
        mb = message.block
        if mb == 'toolbar':
            for e in self.screen.toolbar:
                if e.name == elname:                
                    return e
        else:            
            blnames = message.block.split('@')
            root_block_name = blnames[-1]
            for bl in flatten(self.blocks):
                if bl.name == root_block_name:
                    for blname in blnames[-2::-1]: #reverse for searching nested blocks
                        for c in flatten(bl.value):
                            if c.name == blname and c.type == 'block':
                                bl = c
                                break
                        else:
                            return None                        
                    else:
                        if elname:
                            for c in flatten(bl.value):
                                if c.name == elname:
                                    return c
                        else:
                            return bl
        
    def find_path(self, elem) -> list:        
        def find_in_block(block, elem, path):
            if block == elem:
                return  [block.name, *path]
            for c in flatten(block.value):
                if c == elem:
                    return [c.name, block.name, *path]
                elif c.type == 'block' and (sub_path := find_in_block(c, elem, [block.name, *path])):
                    return sub_path
            return None

        for bl in flatten(self.blocks):        
            if bl == elem:
                return [bl.name]
            
            if sub_path := find_in_block(bl, elem, []):
                return sub_path
        
        if elem in self.screen.toolbar:
            return [elem.name, 'toolbar']

    def prepare_result(self, raw):        
        reload_screen = any(u.type == 'screen' for u in self.changed_units)
        if raw is True or raw == Redesign or reload_screen:            
            self.screen.reload = reload_screen or raw == Redesign                              
            raw = self.screen
        else:
            match raw:
                case None: 
                    if self.changed_units:
                        raw = Message(*self.sorted_changed_units, user = self)
                        if not raw.updates:
                            raw = None
                case Message():
                    if self.changed_units:
                        message_units = [x['data'] for x in raw.updates]
                        self.changed_units.update(message_units)
                        raw.set_updates(self.sorted_changed_units)
                    raw.fill_paths4(self)                                    
                case Unit():
                    self.changed_units.add(raw)
                    raw = Message(*self.sorted_changed_units, user = self) 
                case list() | tuple(): #raw is *unit
                    self.changed_units.update(raw)
                    raw = Message(*self.sorted_changed_units, user = self) 
                case _: ...
                    
        self.changed_units.clear()           
        return raw
    
    def screen_process(self, message):
        screen_change_message = message.screen and self.screen.name != message.screen
        if screen_change_message or message.screen_type:
            for s in self.screens:
                if s.name == message.value:
                    if self.screen_module != s:
                        self.changed_units.add(s.screen)
                        self.screen_module = s   
                        if screen_change_message:
                            break                    
                        if self.voice:
                            self.voice.set_screen(s.screen)       
                            self.voice.start()                              
                        if getattr(s.screen,'prepare', None):
                            s.screen.prepare()
                    return True                     
            else:        
                error = f'Unknown screen name: {message.value}'   
                self.log(error)
                return Error(error)

    async def process(self, message):        
        if screen_result := self.screen_process(message):
            return screen_result
        elif message.voice_type:            
            if not self.voice:
                self.voice = VoiceCom(self)                
            if message.event == 'listen':                
                if message.value:
                    self.voice.start()  
                else:
                    self.voice.stop()
            else:
                return await self.voice.process_string(message.value)            
        else:        
            elem = self.find_element(message)          
            if elem:                          
                return await self.process_element(elem, message)              
            error = f'Element {message.element}@{message.block} does not exist!'
            self.log(error)
            return Error(error)    
            
    async def process_element(self, elem, message):                
        event = message.event         
        handler = self.handlers.get((elem, event), None)
        if handler:
            return await self.eval_handler(handler, elem, message.value)                                                                    
        if hasattr(elem, event):                
            attr = getattr(elem, event)
            if is_callable(attr):
                result = await self.eval_handler(attr, elem, message.value)
                if event in ('complete', 'append', 'get'):                        
                    result = Answer(event, message, result)                
                return result
            #set attribute only for declared properties
            setattr(elem, event, message.value)
        elif event == 'changed':            
            elem.value = message.value                                        
        else:
            error = f"{message.element}@{message.block} doesn't contain '{event}' method type!"
            self.log(error)                     
            return Error(error)
        
    def monitor(self, session, share = None):
        if config.share and session != testdir:
            self.log(f'User is connected, session: {session}, share: {share.session if share else None}', type = 'info')            

    def sync_send(self, obj):                    
        asyncio.run(self.send(obj))
    
    def log(self, message, type = 'error'):        
        scr = self.screen.name if self.screens else 'void'
        message = f"session: {self.session}, screen: {scr}, message: {self.last_message}\n  {message}"
        with logging_lock:
            if type == 'error':
                logging.error(message)
            elif type == 'warning':
                logging.warning(message)    
            else:
                func = logging.getLogger().setLevel
                func(level = logging.INFO)
                logging.info(message)
                func(level = logging.WARNING)

    def init_user():
        """make initial user for autotest and evaluating dbsharing"""
        user = User.type(testdir)
        user.load()    
        #register shared db map once
        user.calc_dbsharing()
        return user

    def calc_dbsharing(self):
        """calc connections db and units"""
        dbshare.clear()
        for module in self.screens:
            screen = module.screen
            for block in flatten(screen.blocks):
                for elem in flatten(block.value):
                    if hasattr(elem, 'id'):
                        dbshare[elem.id][screen.name].append({'element': elem.name, 'block': block.name})                                

    async def sync_dbupdates(self):
        sync_calls = []
        for id, updates in dbupdates.items():
            for update in updates:
                if update:
                    screen2el_bl = dbshare[id]
                    exclude = update.get('exclude', False)
                    for user in Unishare.sessions.values():                
                        if not exclude or user is not self:
                            scr_name = user.screen.name
                            if scr_name in screen2el_bl:
                                for elem_block in screen2el_bl[scr_name]: 
                                    update4user = {**update, **elem_block}
                                    sync_calls.append(user.send(update4user))
        dbupdates.clear()
        await asyncio.gather(*sync_calls)



# Copyright © 2024 UNISI Tech. All rights reserved.
from .units import *
from .common import pretty4, flatten, delete_unit, close_message
from numbers import Number
        
class Block(Unit):    
    def __init__(self, name, *elems, **options):    
        self._mark_changed = None        
        self.name = name        
        self.type = 'block'
        self.value = list(elems)       
        self._user = None 
        self.add(options)  
        if getattr(self,'scaler', False):
            scaler = ContentScaler(elements = lambda: self.scroll_list)
            self.scaler = scaler
            if not self.value:
                self.value = [[scaler]]
            elif isinstance(self.value[0], list):
                self.value[0].append(scaler)
            else:
                self.value[0] = [self.value, scaler]     

        for elem in flatten(self.value):                        
            if hasattr(elem, 'llm'): 
                if elem.llm is True:
                   dependencies = [obj for obj in flatten(self.value) if elem is not obj and obj.type != 'command'] 
                   exactly = False                               
                elif isinstance(elem.llm, list | tuple):
                    dependencies = elem.llm                    
                    exactly = True
                elif isinstance(elem.llm, Unit):                    
                    dependencies = [elem.llm]
                    exactly = True
                elif isinstance(elem.llm, dict):
                    if elem.type != 'table':                        
                        raise AttributeError(f'{elem.name} llm parameter is a dictionary only for tables, not for {elem.type}!')                                                                
                    elem._llm_dependencies = {fld: (deps if isinstance(deps, list | bool) else [deps]) for fld, deps in elem.llm.items()} 
                    elem.llm = True
                    continue
                else:
                    raise AttributeError(f'Invalid llm parameter value for {elem.name} {elem.type}!')
                if dependencies:
                    elem.llm = exactly                
                    for dependency in dependencies:
                        dependency.add_changed_handler(elem.emit)    
                    elem._llm_dependencies = dependencies
                else:
                    elem.llm = None
                    print(f'Empty dependency list for llm calculation for {elem.name} {elem.type}!')
        
        if hasattr(self,'closable'):        
            def close(*_):
                user = self._user if self._user else  Unishare.context_user()
                delete_unit(user.screen.blocks, self.name)                
            self.close = close      

    def __setattr__(self, name, value):      
        super().__setattr__(name, value)
        if name == 'value' and getattr(self, '_user', None):
            self.set_reactivity(self._user)  

    def set_reactivity(self, user, override = False):
        self._user = user
        if user:            
            super().set_reactivity(user, override)
            for elem in flatten(self.value):
                elem.set_reactivity(user)

    @property
    def params(self) -> dict:
        return {el.name : el.value for el in flatten(self.value)}
                
    @property
    def compact_view(self) -> str:
        return ','.join(obj.compact_view for obj in flatten(self.value) if obj.value)

    @property
    def scroll_list(self):            
        return self.value[1] if len(self.value) > 1 and isinstance(self.value[1], (list, tuple)) else []
    
    @scroll_list.setter
    def scroll_list(self, lst):
        self.value = ChangedProxy([self.value[0] if self.value else [], lst], self)
        if hasattr(self,'scaler'):
            sval = self.scaler.value
            if sval != 1:
                self.scaler.value = 1
                self.scaler.changed(self.scaler, sval)                  
        for image in lst:
            image.set_reactivity(self._user)      

    def find(self, elem: Unit | str):
        for e in flatten(self.value):
            if e == elem or e.name == elem:
                return e

class ParamBlock(Block):
    def __init__(self, name, *args, changed = None, row = 3, strict = 'recurse', **params):
        """strict == 'recurse' means to recurse into dict values as an embedded ParamBlock"""
        self._mark_changed = None
        if not args:
            args = [[]]        
        self.name = name        
        self.type = 'block'
        self.value = list(args)
        self._name2elem = {}
        cnt = 0        
        for param, val in params.items():                    
            pretty_name = pretty4(param)            
            match val:
                case True | False:
                    el = Switch(pretty_name, val, changed)
                case str() | int() | float():
                    el = Edit(pretty_name, val, changed)
                case tuple() | list():
                    if len(val) != 2 or not isinstance(val[0], dict):
                        continue
                        #raise ValueError('Composite value has to contain the current value and options value!')
                    options = val[1]
                    if not isinstance(options, list | tuple | dict):
                        raise ValueError('Options value (the second parameter) has to be a list or tuple!')
                    if len(options) == 3 and all(map(lambda e: isinstance(e, Number), options)):
                        el = Range(pretty_name, val[0], changed, options = options)
                    elif isinstance(options, list | tuple):
                        el = Select(pretty_name, val[0], changed, options = options, type = 'select')
                    else: 
                        el = Tree(pretty_name, val[0], changed, options = options)
                case _:
                    if strict == 'recurse' and isinstance(val, dict):
                        pb = ParamBlock(pretty_name, changed = changed, strict = strict, row = row, **val)
                        self.value.append(pb)
                        self._name2elem[param] = pb
                        cnt = 0                        
                    elif strict:
                        raise ValueError(f'The {param} value {val} is not supported. Look at ParamBlock documentation!')                    
                    continue
                    
            self._name2elem[param] = el
            if cnt % row == 0:
                block = []
                self.value.append(block)
            cnt += 1
            block.append(el)
        
    @property
    def params(self) -> dict:        
        return {name: el.params for name, el in self._name2elem.items()}

class Dialog:  
    def __init__(self, question, callback, *content, commands = ['Ok','Cancel'],
            icon = 'not_listed_location'):        
        self.type = 'dialog'         
        self.name = question
        self.changed = callback          
        buttons = [Button(name, color = 'secondary', width = 80, close = True) for name in commands]
        for button in buttons:
            button.changed = self.dialog_command_handler      
        buttons[0].color = 'primary' 
        buttons[0].space = True        
        self.icon = icon
        self.value = [[], *content, buttons] if content else buttons        

    async def dialog_command_handler(self, button, _):        
        if user := Unishare.context_user():
            user.active_dialog = None
            await user.send(TypeMessage('action', 'close'))
            return await call_anysync(self.changed, self, button.name)        

class Screen(Unit):
    def __init__(self, name):
        self._mark_changed = None        
        self.name = name                
        self.type = 'screen'                                  

    def set_reactivity(self, user, override = False):
        super().set_reactivity(user, override)
        for block in flatten(self.blocks):
            block.set_reactivity(user, override)
        for elem in self.toolbar:
            elem.set_reactivity(user, override)



from data import table
from unisi import *
from blocks.tblock import eblock
name = 'Panda & params'
order = 1

zoo_table = Table('Zoo Table (panda table)', panda = table)

def get_params(button, _):
    return Info(str(block.params))

def block_changed(element, value): 
    element.value = value
    return Info(f"{element.name} is {value}.")

block = ParamBlock('System parameters', Button('Show server params', get_params), 
        changed = block_changed, strict = 'recurse',
    per_device_eval_batch_size=16,
    num_train_epochs=10, 
    warmup_ratio=0.1, 
    inference = dict(
        logging_steps= (10,[1,20,1]), 
        device = ('gpu', ['cpu', 'gpu']),
        load_best = True))

def html_handler(unit, event):
    return Info(event)

html = HTML('HTML',  '<button>Click me</button> <a href="#">Link</a> <input type="text">') #, html_handler)

html_block = Block('Block with HTML', [], html, zoo_table)

blocks = [[eblock,block], html_block]




from unisi import *
name = 'Main'

table = Table('Persons', headers = ['Name', 'Date of birth','Occupation'], 
                llm = {'Date of birth': 'Name', 'Occupation': True},
    rows = [['Michael Jackson', None, None], ['Ronald Reagan', None, None]])

block1 = Block('Relations',Button('Calculate the selected', table.emit), table)

ename = Edit('Name')

ebirth = Edit('Date of birth', llm = True)

block2 = Block('Enter any name', [ename, Button('Calculate birth date', ebirth.emit)], [ebirth, Edit('Occupation', llm = ename)]) 

ecountry = Edit('Country', "France")

ecountry_info = Edit('Country info')

tcapitals = Table('Europe', headers = ['Country','Capital'], tools = False)

topic = 'Managing personal notes' 
etopic = Edit('Topic', topic)
extext = TextArea('Main points')
equestions = TextArea(f'Questions for the topic')

async def geo_code():    
    country = ecountry.value.strip()
    if country:
        await user.progress(f'Information about {country}...')
        country_info = await Q("Provide information about {country}.", dict(capital = str, population = int, currency = str))
        ecountry_info.value = str(country_info)
    await user.progress(f'Information about Europe...')
    country2capital = await Q("What are the capitals of the European countries?", dict[str, str])        
    tcapitals.rows = [[country, capital] for country, capital in country2capital.items()]    
        
text = "Alpha Centauri A, also known as Rigil Kentaurus, is the principal member, or primary,\
 of the binary system. It is a solar-like main-sequence star with a similar yellowish colour, whose stellar \
 classification is spectral type G2-V; it is about 10% more massive than the Sun, with a radius about 22% larger.",
async def extract_info(*_):
    key_points = await Q("Extract a list of key points from the following text: {text}. ", list[str], text=text) #and return them as a list
    etext = ''
    for i, key_point in enumerate(key_points):
        etext = f"{etext}{i+1}. {key_point}\n"
    extext.value = etext
    await parallel_execution()
    
async def parallel_execution():        
    questions = await Q("Suggest a list of 3 questions about {topic}.", list[str], topic=topic)
    Qs = [Q(question, str) for question in questions]
    results = await asyncio.gather(*Qs) 
    etext = ''
    i = 1
    for question, result in zip(questions, results):
        etext = f"==={etext}{i}. {question}\n  => {result}\n"
        i += 1
    equestions.value = etext

async def test(*_):
    await user.progress('Calculating...')
    await geo_code()

button = Button("Run", test)
tblock = Block("Geo calculations", [ecountry, button], ecountry_info, tcapitals, width=400)

eblock = Block('Text operations', [etopic, Button('Extract info', extract_info)], extext, equestions)

blocks = [[block2, block1], [tblock, eblock]]


# Copyright © 2024 UNISI Tech. All rights reserved.
import kuzu, shutil, os, re
from datetime import date, datetime
from cymple import QueryBuilder as qb
from cymple.typedefs import Properties
from .common import get_default_args
from .dbunits import Dblist

def equal_fields_dicts(dict1, dict2):
    return dict1.keys() == dict2.keys() and all(dict1[key].lower() == dict2[key].lower() for key in dict1)

def is_modifying_query(cypher_query):
    query = cypher_query.lower()
    modifying_pattern = r'\b(create|delete|detach\s+delete|set|merge|remove|call\s+\w+\s+yield|foreach)\b'        
    return re.search(modifying_pattern, query)

def query_offset(id, offset):
    return qb().match().node(id, 'a').where('a.ID','=',offset)  

def kuzu_data_type(value):
    match value:
        case bool():
            return "BOOLEAN"
        case int():        
            return "INT64"
        case float():
            return "DOUBLE"
        case str():
            return "STRING"
        case datetime():
            return "TIMESTAMP"        
        case date():
            return "DATE"
        case bytes():
            return "BLOB"          
        case list() | tuple():
            return "LIST"
        case _:
            return ""

number_types = ["DOUBLE", "INT64"]

def dict_to_cypher_set(properties, alias = 'a'):
    set_items = []
    for key, value in properties.items():
        if isinstance(value, str):    
            set_items.append(f"{alias}.{key} = '{value}'")
        else:
            set_items.append(f"{alias}.{key} = {value}")        
    return "SET " + ", ".join(set_items)

class Database:
    tables = {} #id -> Dbtable
    def __init__(self, dbpath, message_logger = print) -> None:    
        self.db = kuzu.Database(dbpath)
        self.conn = kuzu.Connection(self.db)
        self.message_logger = message_logger
        self.table_params = get_default_args(self.get_table)
    
    def execute(self, query_str, ignore_exception = False):
        query_str = str(query_str)
        """ if not query_str.endswith(';'):
            query_str += ';'
        print(query_str) """
        try:
            result = self.conn.execute(query_str)            
        except Exception as e:
            if not ignore_exception:
                self.message_logger(e)
            return None
        return True if result is None else result            
        
    def delete(dir_path):
        if os.path.exists(dir_path):
            # Remove the directory and all its contents
            shutil.rmtree(dir_path)            

    @property
    def table_names(self):
        return self.conn._get_node_table_names()
    
    def get_table_fields(self, table_name, remove_id = True) -> None | dict:                            
        result = self.qlist(f"CALL table_info('{table_name}') RETURN *;", ignore_exception = True)
        if result is not None:
            return {info[1]: info[2] for info in result if not remove_id or info[1] != 'ID'}

    def delete_table(self, table_name):        
        return self.execute( f'DROP TABLE {table_name};')
            
    def get_table(self, id = None, limit = 100, headers = None, rows = None, fields = None):        
        if id:            
            if rows and fields is None:
                if not headers:
                    self.message_logger(f'headers are not defined!')
                    return None
                types = [None] * len(headers)                
                for row in rows:
                    for j, cell in enumerate(row):
                        if cell is not None:
                            ktype = kuzu_data_type(cell)
                            if ktype:
                                if types[j] is None:
                                    types[j] = ktype
                                elif types[j] != ktype:
                                    if types[j] in number_types and ktype in number_types:
                                        types[j] = "DOUBLE"
                                    else:
                                        self.message_logger(f'Conflict types for {id} table in {j} column: {types[j], ktype}!')
                                        return None                    
                if None in types:
                    index = types.index(None)
                    self.message_logger(f'Rows data doesnt contain allowed values for {headers[index]} column!')
                    return None
                fields = {headers[i]: type for i, type in enumerate(types)}
                        
            if (table_fields := self.get_table_fields(id)) is not None:
                if not equal_fields_dicts(table_fields, fields):                    
                    if self.delete_table(id):      
                        self.message_logger(f'Node table {id} was deleted because of fields contradiction!', 'warning')          
                else:
                    return self.tables.get(id) or Dbtable(id, self, limit, table_fields)
                        
        return self.create_table(id, fields, limit, rows)   

    def get_table_params(self, params):    
        return {k: v for k, v in params.items() if k in self.table_params}   

    def set_db_list(self, gui_table):
        table = self.get_table(**self.get_table_params(gui_table.__dict__))
        tlst = table.list
        gui_table.rows = tlst        
                    
    def create_table(self, id, fields : dict, limit = 100, rows = None):                
        specs = ','.join(f'{prop} {type}' for prop, type in fields.items())
        query = f"CREATE NODE TABLE {id}({specs},ID SERIAL, PRIMARY KEY(ID))"
        self.execute(query)
        table = Dbtable(id, self, limit, fields)
        if rows:
            table.list.extend(rows)
        return table
    
    def update_row(self, table_id, row_id, props, in_node = True):        
        set_props = dict_to_cypher_set(props)
        query = f'MATCH (a: {table_id}) WHERE a.ID = {row_id} {set_props}' if in_node else\
            f'MATCH ()-[a: {table_id}]->() WHERE a.ID = {row_id} {set_props}'
        return self.execute(query) 
        
    def qlist(self, query, func = None, ignore_exception = False):                
        if answer := self.execute(query, ignore_exception):            
            result = []
            while answer.has_next():
                value = answer.get_next()
                result.append(func(value) if func else value)
            return result
    
    def qiter(self, query, func = None, ignore_exception = False):        
        answer = self.execute(query, ignore_exception)
        while answer.has_next():
            value = answer.get_next()
            yield func(value) if func else value        
            
class Dbtable:
    def __init__(self, id, db, limit = 100, table_fields = None) -> None:        
        self.db = db
        db.tables[id] = self
        self.id = id              
        self.table_fields = table_fields
        self.limit = limit    
        self.node_columns = list(db.conn._get_node_property_names(id).keys())[:-1]            
        self.init_list()          

    @property
    def rel_table_names(self):
        return self.db.conn._get_rel_table_names() 
    
    def default_index_name2(self, link_table):
        return f'{self.id}2{link_table}'
    
    def calc_linked_rows(self,  index_name, link_ids, include_rels = False, search = ''):
        condition = f'b.ID in {link_ids}'
        rel_info = ', r.*' if include_rels else ''
        query = f"""
            MATCH (a:{self.id})-[r:{index_name}]->(b:User)
            WHERE {condition}
            RETURN a.*{rel_info}
            ORDER BY a.ID ASC
        """        
        lst = self.db.qlist(query)         
        return Dblist(self, cache = lst)
        
    def get_rel_fields2(self, tname, fields : dict = None, relname = None):
        """return name of link table and fields and its fields dict"""
        if not relname:
            relname = self.default_index_name2(tname)
        rel_table_fields = self.db.get_table_fields(relname)
        if isinstance(rel_table_fields, dict):
            if isinstance(fields, dict):
                if equal_fields_dicts(rel_table_fields, fields):
                    return relname, rel_table_fields
                else:
                    self.db.delete_table(relname)            
            else:
                fields = rel_table_fields
        elif fields is None:
            fields = {}

        if not any(info['name']  == relname for info in self.rel_table_names):                        
            fprops = ''.join(f', {field} {type}' for field, type in fields.items()) if fields else ''
            fprops += ', ID SERIAL'
            query = f"CREATE REL TABLE {relname}(FROM {self.id} TO {tname} {fprops})"
            self.db.execute(query)
            self.rel_table_names.append({'name' : relname})
        return relname, fields
    
    def add_link(self, snode_id, link_table, tnode_id, link_fields = None, link_index_name = None):
        """return added link"""
        if link_index_name is None:
            link_index_name = self.default_index_name2(link_table)
        if link_fields is None:
            link_fields = {}
        query = f"""
            MATCH (a:{self.id}), (b:{link_table})
            WHERE a.ID = {snode_id} AND b.ID = {tnode_id}
            CREATE (a)-[r:{link_index_name} {{{Properties(link_fields)}}}]->(b)            
            RETURN r.*
            """
        lst = self.db.qlist(query)
        return lst[0]
    
    def add_links(self, link_table, snode_ids : iter, tnode_id, link_index_name = None):
        result = []
        for id in snode_ids:
            result.append(self.add_link(id, link_table, tnode_id, link_index_name = link_index_name))
        return result
    
    def delete_link(self, link_table_id, link_id, index_name = None):
        if not index_name:
            index_name = self.default_index_name2(link_table_id)
        query = f"""
        MATCH (:{self.id})-[r:{index_name}]->(:{link_table_id})
        WHERE r.ID = {link_id}
        DELETE r
        """
        self.db.execute(query)

    def delete_links(self, link_table_id, link_node_id = None, source_ids = None, link_ids = None, index_name = None):
        if not index_name:
            index_name = self.default_index_name2(link_table_id)
        
        if link_ids:
            condition = f'r.ID in {link_ids}'            
        else:
            if not isinstance(source_ids, list):
                source_ids = list(source_ids)                        
            condition = f'(a.ID in {source_ids}) AND b.ID = {link_node_id}'
        query = f"""
        MATCH (a:{self.id})-[r:{index_name}]->(b:{link_table_id})
        WHERE {condition}
        DELETE r
        """
        self.db.execute(query)

    def init_list(self):                
        list = self.read_rows(limit = self.limit)
        length = len(list)
        #possibly the table has more rows        
        if length == self.limit:            
            ql = self.db.qlist(f"MATCH (n:{self.id}) RETURN count(n)")
            self.length = ql[0][0]
        else:
            self.length = length
        self.list = Dblist(self, list)
        
    def read_rows(self, skip = 0, limit = 0):
        query = qb().match().node(self.id, 'a').return_literal('a.*').order_by('a.ID')
        if skip:
            query = query.skip(skip)
        query = query.limit(limit if limit else self.limit)
        return self.db.qlist(query)    

    def assign_row(self, row_array):
        return self.db.update_row(self.id, row_array[-1], 
            {name : value for name, value in zip(self.node_columns, row_array)})

    def delete_row(self, id):
        query = query_offset(self.id, id)
        self.length -= 1
        return self.db.execute(query.detach_delete('a'))
    
    def delete_rows(self, ids):
        condition = f'a.ID in {ids}'
        query = f"""
        MATCH (a:{self.id})
        WHERE {condition}
        DELETE a
        """
        return self.db.execute(query)
    
    def clear(self, detach = False):
        query = f'MATCH (a:{self.id})'
        if detach:
            query += ' DETACH DELETE a'
        else:
            query += ' DELETE a'
        self.length = 0
        return self.db.execute(query)
    
    def append_row(self, row):
        """row can be list or dict, returns new row"""        
        if isinstance(row, list):
            props = {name: value for name, value in zip(self.node_columns, row) if value is not None}
                      
        answer = self.db.execute(qb().create().node(self.id, 'a', props).return_literal('a.*'))        
        if answer and answer.has_next():                        
            self.length += 1
            return answer.get_next()        
    
    def append_rows(self, rows):
        """row can be list or dict"""
        rows_arr = []
        for row in rows:            
            row = {name: value for name, value in zip(self.node_columns, row)} if not isinstance(row, dict) else row
            srow = f' {{{Properties(row).to_str()}}}'
            rows_arr.append(srow)      
        rows_arr = ','.join(rows_arr)      
        
        query = (qb().with_(f'[{rows_arr}] AS rows')
            .unwind('rows AS row')
            .create()
            .node(self.id, 'n', {p: f'row.{p}' for p in self.node_columns}, escape=False)
            .return_literal('n.*'))
        
        self.length += len(rows)        
        return self.db.qlist(query)  




# Copyright © 2024 UNISI Tech. All rights reserved.
import multiprocessing, time, asyncio, logging, inspect
from .utils import start_logging
from config import froze_time, monitor_tick, profile, pool

def write_string_to(shared_array, input_string):    
    input_bytes = input_string.encode()    
    shared_array[:len(input_bytes)] = input_bytes

def read_string_from(shared_array):
    return shared_array[:].decode().rstrip('\x00')

_multiprocessing_pool = None

def multiprocessing_pool():
    global _multiprocessing_pool
    if not _multiprocessing_pool:
        _multiprocessing_pool = multiprocessing.Pool(pool)
    return _multiprocessing_pool

async def run_external_process(long_running_task, *args, progress_callback = None, **kwargs):
    if progress_callback:
        if args[-1] is None:
            queue = multiprocessing.Manager().Queue()
            args = *args[:-1], queue
        else:
            queue = args[-1] 
                        
    result = multiprocessing_pool().apply_async(long_running_task, args, kwargs)
    if progress_callback:
        while not result.ready() or not queue.empty():            
            message = queue.get()
            if message is None:
                break
            await asyncio.gather(progress_callback(message), asyncio.sleep(monitor_tick))            
    return result.get()

logging_lock = multiprocessing.Lock()

splitter = '~'

def monitor_process(monitor_shared_arr):            
    timer = None
    session_status = {}    
    sname = None    
    start_logging()
    while True:
        #Wait for data in the shared array
        while monitor_shared_arr[0] == b'\x00':
            time.sleep(monitor_tick)  
            if timer is not None:
                timer -= monitor_tick                
                if timer < 0:
                    timer = None                    
                    arr = list(session_status.items())
                    arr.sort(key = lambda s: s[1][1], reverse=True)
                    ct = time.time()
                    message = "Hangout is detected! Sessions in a queue and time waiting:" +\
                        ''.join(f'\n  {s[0]}, {s[1][0]}, {ct - s[1][1]} s' for s in arr)    
                    with logging_lock:
                        logging.warning(message)                    
                    timer = None
        # Read and process the data
        status = read_string_from(monitor_shared_arr).split(splitter)
        #free
        monitor_shared_arr[0] = b'\x00'
        sname = status[1]  
        match status[0]:
            case '+' | 'e': #exit external process                             
                session_status[sname] = [status[2], time.time()]
                timer = froze_time
            case '-':    
                event, tstart = session_status.get(sname, (None, 0))
                if event:                                        
                    duration = time.time() - tstart
                    if profile and duration > profile:
                        with logging_lock:
                            logging.warning(f'Event handler {event} was executed for {duration} seconds!')
                    del session_status[sname] 
                    timer = None
            case 'p': #call external process
                session_status[sname] = [status[2], time.time()]
                timer = None
                            
if froze_time or profile: 
    # Create a shared memory array
    monitor_shared_arr = multiprocessing.Array('c', 200)  
    monitor_shared_arr[0] != b'\x00'
    
    async def notify_monitor(status, session, event):
        s = f'{status}{splitter}{session}{splitter}{event}'        
        # Wait for the shared array to be empty
        while monitor_shared_arr[0] != b'\x00':
            await asyncio.sleep(monitor_tick)
        write_string_to(monitor_shared_arr, s)

    monitor_process = multiprocessing.Process(target=monitor_process, args=(monitor_shared_arr,))
    monitor_process.start()
else:
    notify_monitor = None
    


        


import pandas as pd, time
table = pd.read_csv('zoo.csv')

def long_function(ticks, queue):
    queue.put('Run process')
    for i in range(ticks):        
        time.sleep(0.04)
        queue.put(f'{i} tick')
    return 5



from abc import ABC


class Error(ABC):
    expected = None
    received = None

    template = 'Expected: <{e}>, received: <{r}>'

    def __init__(self, expected, received):
        self.expected = expected
        self.received = received

    @property
    def message(self):
        msg = self.template.format(e=self.expected, r=self.received)
        return msg

    def explain(self):
        return {
            '_message': self.message,
            '_expected': self.expected,
            '_received': self.received,
        }


class TypesNotEqual(Error):
    template = 'Types not equal. Expected: <{e}>, received: <{r}>'

    def __init__(self, e, a):
        e = type(e).__name__
        a = type(a).__name__
        super().__init__(e, a)


class ValuesNotEqual(Error):
    template = 'Values not equal. Expected: <{e}>, received: <{r}>'


class KeyNotExist(Error):
    template = 'Key does not exist. Expected: <{e}>'


class LengthsNotEqual(Error):
    template = 'Lengths not equal. Expected <{e}>, received: <{r}>'


class ValueNotFound(Error):
    template = 'Value not found. Expected <{e}>'


class UnexpectedKey(Error):
    template = 'Unexpected key. Received: <{r}>'



# Copyright © 2024 UNISI Tech. All rights reserved.
from .common import *
from .llmrag import get_property

class ChangedProxy:
    MODIFYING_METHODS = {
        'append', 'extend', 'insert', 'remove', 'pop', 'clear', 'sort', 'reverse',
        'update', 'popitem', 'setdefault', '__setitem__', '__delitem__'
    }
    def __init__(self, obj, unit):
        self._obj = obj
        self._unit = unit
    
    def __getattribute__(self, name):        
        if name == '_obj' or name == '_unit' or name == '__getstate__':
            return super().__getattribute__(name)
        obj = super().__getattribute__('_obj')
        value = getattr(obj, name)  
        if isinstance(value, ChangedProxy):
            value = value._obj
        if name in ChangedProxy.MODIFYING_METHODS:
            super().__getattribute__('_unit')._mark_changed()
        elif not callable(value) and not isinstance(value, atomics):
            return ChangedProxy(value, self)
        return value
    
    def __setattr__(self, name, value):                
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            self._obj.__setattr__(name, value) 
            self._unit._mark_changed()

    def __setitem__(self, key, value):
        self._obj[key] = value
        self._unit._mark_changed ()
       
    def __getitem__(self, key):
        value = self._obj[key]    
        if not callable(value) and not isinstance(value, atomics):
            value = ChangedProxy(value, self._unit)
        return value
    
    def __eq__(self, other):
        return self._obj.__eq__(other._obj) if isinstance(other, ChangedProxy) else self._obj.__eq__(other)

    def __delitem__(self, key):
        del self._obj[key]
        self._unit._mark_changed ()

    def __iter__(self):
        for item in self._obj:
            if not callable(item) and not isinstance(item, atomics):
                item = ChangedProxy(item, self._unit)
            yield item        

    def __len__(self):
        try:
            return len(self._obj)
        except TypeError:        
            return 0  
        
    def __hash__(self):
        return hash(self._obj)
        
    def __iadd__(self, other):  
        if isinstance(self._obj, list):
            self.extend(other)  
            return self  # Important: __iadd__ must return self
        
        raise TypeError(f"Unsupported operand type for += with '{type(self._obj).__name__}'")
            
    def __getstate__(self):     
        return self._obj

           
class Unit:    
    action_list = set(['complete', 'update', 'changed','delete','append', 'modify'])
    def __init__(self, name, *args, **kwargs):                
        self._mark_changed =  None
        self.name = name
        la = len(args)
        if la:
            self.value = args[0]
        if la > 1:
            self.changed = args[1]                    
        self.add(kwargs)

    def specific_changed_register(self, property, value) -> bool:
        """ addtional actions when changed, return False if not changed"""
        return not property or not property.startswith('_')
        
    def set_reactivity(self, user, override = False):        
        changed_call = self._mark_changed
        
        if not hasattr(self, 'id') and (override or not changed_call): 
            self.__dict__.update({property : ChangedProxy(value, self)  for property, value in self.__dict__.items() 
                if property[0] != '_' and not isinstance(value, atomics) and not callable(value)})                    
                    
            def changed_call(property = None, value = None):
                if self.specific_changed_register(property, value):
                    user.register_changed_unit(self, property, value)            
        super().__setattr__('_mark_changed', changed_call)                    

    def add(self, kwargs):              
        for key, value in kwargs.items():
            setattr(self, key, value)   

    def __setattr__(self, name, value):      
        #it is correct condition order 
        if name[0] != "_" and self._mark_changed:            
            self._mark_changed(name, value)
            if not callable(value) and not isinstance(value, atomics):
                value = ChangedProxy(value, self)
        super().__setattr__(name, value)        

    def mutate(self, obj):
        if self is not obj:
            self.__dict__.clear()
            for key, value in obj.__dict__.items():
                setattr(self, key, value)
            if self._mark_changed:
                self._mark_changed()
    
    def accept(self, value):
        if hasattr(self, 'changed'):
            self.changed(self, value)
        else:
            self.value = value

    def delattr(self, attr):
        if hasattr(self, attr): 
            delattr(self, attr)

    def __eq__(self, other):
        return super().__eq__(other._obj) if isinstance(other, ChangedProxy) else super().__eq__(other)
    
    def __hash__(self):
        return super().__hash__()
    
    @property
    def params(self):
        return self.value

    @property
    def compact_view(self) -> str:
        """reduce for external (llm) using if required"""
        return f'{self.name} : {self.value}'
    
    @property
    def type_value(self):
        """return python system type if value not 'date' type"""
        return type(self.value) if self.type != 'date' else 'date'
    
    async def emit(self, *_ ):        
        """calcute value by system llm, can be used as a handler"""
        if Unishare.llm_model and (exactly := getattr(self, 'llm', None)) is not None:        
            elems = [e.compact_view for e in self._llm_dependencies if e.value != '' and e.value is not None]
            #exactly is requirment that all elements have to have valid value
            if not exactly or len(elems) == len(self._llm_dependencies):
                context = ','.join(elems)    
                self.value = await get_property(self.name, context, self.type_value, options = getattr(self, 'options', None))
                return self
            
    def add_changed_handler(self, handler):
        changed_handler = getattr(self, 'changed', None)
        if not changed_handler:
            def changed_handler(obj, value):
                obj.value = value
        self.changed = compose_handlers(changed_handler, handler) 
    
    def __getstate__(self):         
        return {n: (True if n in Unit.action_list else v) for n, v in self.__dict__.items() if n[0] != '_'}

    def __str__(self):
        return f'{type(self).__name__}({self.name})'
    
    def __repr__(self):
        return f'{type(self).__name__}({self.name})'
    
atomics = (int, float, complex, bool, str, bytes, ChangedProxy, type(None), Unit)

Line = Unit("__Line__", type = 'line')

def smart_complete(lst, min_input_length = 0, max_output_length = 20):
    di = {it: it.lower() for it in lst}
    def complete(_, ustr):
        if len(ustr) < min_input_length:
            return []
        ustr = ustr.lower()
        arr = [(itlow.find(ustr), it, itlow) for it, itlow in di.items() if itlow.find(ustr) != -1]
        arr.sort(key = lambda e: (e[0], e[2]))
        if len(arr) > max_output_length:
            arr = arr[: max_output_length]
        return [e[1] for e in arr]
    return complete

class Edit(Unit):
    def __init__(self, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)        
        self.x = 0
        has_value = hasattr(self,'value')
        if 'type' not in kwargs:            
            if has_value:
                type_value = type(self.value)
                if type_value == int or type_value == float:
                    self.type = 'number'
                    return
            self.type = 'string'
        if not has_value:
            self.value = '' if self.type != 'number' else 0

class Text(Unit):
    def __init__(self, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.value = self.name
        self.type = 'string'
        self.edit = False       

class Range(Unit):
    def __init__(self, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)    
        if not hasattr(self, 'value'):
            self.value = 1.0    
        self.type = 'range'                
        if 'options' not in kwargs:
            self.options = [self.value - 10, self.value + 10, 1]

class ContentScaler(Range):
    def __init__(self, *args, **kwargs):
        name = args[0] if args else 'Scale content'        
        super().__init__(name, *args, **kwargs)                
        if 'options' not in kwargs:
            self.options = [0.25, 3.0, 0.25]
        self.changed = self.scaler
        
    def scaler(self, _, val):
        prev = self.value
        elements = self.elements() 
        self.value = val
        if elements:
            prev /= val
            for element in elements:
                element.width /= prev
                element.height /= prev            
            return elements

class Button(Unit):
    def __init__(self, name, handler = None, **kwargs):
        self._mark_changed =  None
        self.name = name
        self.value = None
        self.add(kwargs)
        if not hasattr(self, 'type'):
            self.type = 'command'
        if handler:
            self.changed = handler
            
def CameraButton(name, handler = None, **kwargs):    
    kwargs['type'] = 'camera'
    return Button(name, handler, **kwargs)
        
def UploadButton(name, handler = None,**kwargs):    
    kwargs['type'] = 'uploader'
    if 'width' not in kwargs:
        kwargs['width'] = 250.0                  
    return Button(name, handler, **kwargs)

class Image(Unit):
    '''name is file name or url, label is optional text to draw on the image'''
    def __init__(self, name, value = False, handler = None, label = '', width = 300, **kwargs):
        super().__init__(name, [], **kwargs)
        self.value = value
        if handler:
            self.changed = handler
        self.label = label
        self.type='image'        
        self.width = width        
        if not hasattr(self,'url'):
            self.url = self.name
        #mask full win path from Chrome detector
        if self.url[1] == ':': 
            self.url = f'/{self.url}'

class Video(Unit):
    '''has to contain fragments parameter, value like 
        { "currentTime": 5.123, "isVideoPlaying": False, "sound": False }'''
    def __init__(self,name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.type = 'video'        
        set_defaults(self, {'fragments': []})

class Chart(Unit):
    '''echart chart with 'echart option' parameter as value, handler is optional'''
    def __init__(self,name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.type = 'chart'                
        self.option = self.value if hasattr(self, 'value') else {}
        self.value = None
                
class Switch(Unit):
    def __init__(self,name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)        
        set_defaults(self,{'value': False, 'type': 'switch'})

class Select(Unit):
    def __init__(self,name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        set_defaults(self,{'options': [], 'value': None})
        if not hasattr(self, 'type'):
            self.type = 'select' if len(self.options) > 3 else 'radio'        

class Tree(Unit):
    def __init__(self,name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)                 
        set_defaults(self,dict(options = [], value = None, type = 'tree'))
        
class TextArea(Unit):
    def __init__(self,name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.x = 0
        self.type = 'text' 

class HTML(Unit):
    def __init__(self,name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)        
        self.type = 'html' 
                     
        


# Copyright © 2024 UNISI Tech. All rights reserved.
from .common import Unishare
from langchain_groq import ChatGroq
from langchain_openai import ChatOpenAI
from langchain_mistralai import ChatMistralAI
from langchain_xai import ChatXAI
from langchain_google_genai import (
    ChatGoogleGenerativeAI,
    HarmBlockThreshold,
    HarmCategory,
)
from datetime import datetime
import os, inspect, re, json
from typing import get_origin, get_args, Any, Union

class QueryCache:
    ITEM_SEPARATOR = "§¶†‡◊•→±"
    ENTRY_SEPARATOR = "€£¥¢≠≈∆√"    
    
    def __init__(self, file_path):
        self.file_path = file_path
        self.cache = {}
        self._load_cache()

    def _load_cache(self):
        if os.path.exists(self.file_path):
            with open(self.file_path, 'r', encoding='utf-8') as f:
                content = f.read()
                entries = content.split(self.ENTRY_SEPARATOR)
                for entry in entries:
                    if not entry.strip():
                        continue
                    parts = entry.split(self.ITEM_SEPARATOR)
                    if len(parts) == 2:
                        query, result = parts
                        self.cache[query] = result

    def get(self, query):
        return self.cache.get(query)

    def set(self, query, result):        
        entry_data = f"{query}{self.ITEM_SEPARATOR}{result}"
        prepend_separator = bool(self.cache)
        
        try:
            with open(self.file_path, 'a', encoding='utf-8') as f:
                if prepend_separator:
                    f.write(self.ENTRY_SEPARATOR)
                f.write(entry_data)
            
            self.cache[query] = result
            return True
        except Exception as e:
            print(f"Error caching result: {e}")
            return False

def jstype(type_value):        
    if isinstance(type_value, type):         
        if type_value == int:
            return 'integer'
        elif type_value == float:
            return 'number'
        elif type_value == bool:
            return 'boolean'
        elif type_value == str:
            return 'string'
        elif type_value == dict:            
            return 'object'
        elif type_value == list:            
            return 'array'
        else:
            origin = get_origin(type_value) 
            args = get_args(type_value)
            if origin == list:
                return f'array of {jstype(args[0])} '
            elif origin == dict:
                return f'object of {jstype(args[0])} to {jstype(args[1])} structure.'
            else:
                return 'string'  
    else: 
        match type_value:
            case str():
                jtype = 'string'
            case int():
                jtype = 'integer'
            case float():
                jtype = 'number'
            case bool():
                jtype = 'boolean'
            case dict():
                if type_value:
                    ptypes = ','.join(f'"{k}": "[Type: {jstype(v)}]"' for k, v in type_value.items())
                    jtype = f'object with {{{ptypes}}} structure'
                else:
                    jtype = 'object'
            case list():  
                jtype = 'array'
            case _:
                jtype = 'string' 
    return jtype

def is_type(variable, expected_type):
    """
    Check if the variable matches the expected type hint.
    """
    # Handle explicit mapping of expected keys to types: {'name': str, 'age': int}
    if isinstance(expected_type, dict):
        if not isinstance(variable, dict):
            return False
        for key_pattern, sub_type in expected_type.items():
            # direct key match
            if key_pattern in variable:
                if not is_type(variable[key_pattern], sub_type):
                    return False
                continue            
        return True

    # Handle typing hints (e.g., List[int], Dict[str,int], Union[..., ...], Optional[T], Tuple[T,...])
    origin = get_origin(expected_type)
    args = get_args(expected_type)

    # No typing origin: expected_type may be a bare type or typing.Any
    if origin is None:
        # typing.Any
        if expected_type is Any:
            return True
        # None as expected type
        if expected_type is None:
            return variable is None
        # expected_type could be a builtin type or a tuple of types
        if isinstance(expected_type, type) or isinstance(expected_type, tuple):
            return isinstance(variable, expected_type)
        # fallback: try isinstance; for other unexpected cases, be permissive
        try:
            return isinstance(variable, expected_type)
        except Exception:
            return False

    # Handle Union / Optional
    if origin is Union:
        return any(is_type(variable, arg) for arg in args)

    # Handle List[...] and Set[...]
    if origin in (list, set):
        expected_container = list if origin is list else set
        if not isinstance(variable, expected_container):
            return False
        if not args:
            return True
        return all(is_type(item, args[0]) for item in variable)

    # Handle Tuple[...] or tuple
    if origin is tuple:
        if not isinstance(variable, tuple):
            return False
        if not args:
            return True
        # Homogeneous tuple: Tuple[T, ...]
        if len(args) == 2 and args[1] is Ellipsis:
            return all(is_type(item, args[0]) for item in variable)
        # Fixed-length tuple: Tuple[T1, T2, ...]
        if len(variable) != len(args):
            return False
        return all(is_type(item, typ) for item, typ in zip(variable, args))

    # Handle Dict[K, V]
    if origin is dict:
        if not isinstance(variable, dict):
            return False
        if not args:
            return True
        key_type, val_type = args
        return all(is_type(k, key_type) and is_type(v, val_type) for k, v in variable.items())

    # Handle Literal[...] (available in typing)
    try:
        from typing import Literal
        if origin is Literal:
            return any(variable == lit for lit in args)
    except Exception:
        pass

    # Fallback: check using isinstance against the origin if possible
    try:
        return isinstance(variable, origin)
    except Exception:
        return False

def remove_comments(json_str):
    # Regular expression to remove single-line comments (// ...)
    json_str = re.sub(r'//.*', '', json_str)
    # Regular expression to remove multi-line comments (/* ... */)
    json_str = re.sub(r'/\*.*?\*/', '', json_str, flags=re.DOTALL)
    return json_str

def Q(str_prompt, type_value = str, blank = True, extend = True, format = True, **format_model):
    """returns LLM async call for a question, `extend = True` adds system prompt, 
     'identity' in format_model can be used to set the assistant identity"""    
    llm = Unishare.llm_model    
    if format and '{' in str_prompt:
        caller_frame = inspect.currentframe().f_back            
        format_model = caller_frame.f_locals | format_model if format_model else caller_frame.f_locals
        str_prompt = str_prompt.format(**format_model) 
    if extend: 
        if type_value is not None:          
            jtype = jstype(type_value)
            format = " dd/mm/yyyy string" if type_value == 'date' else f'a JSON {jtype}' if jtype != 'string' else jtype      
            str_prompt = f" Output STRONGLY in format {format}. DO NOT OUTPUT ANY COMMENTARY." + str_prompt         
        str_prompt = format_model.get('identity', 'You are an intelligent and extremely smart assistant.') + str_prompt
    async def f():            
        if Unishare.llm_cache:            
            if content := Unishare.llm_cache.get(str_prompt):
                pass
            else:
                io = await llm.ainvoke(str_prompt)
                content = io.content
                Unishare.llm_cache.set(str_prompt, content)
        else:
            io = await llm.ainvoke(str_prompt)
            content = io.content
        js = content.strip().strip('`').replace('json', '')                      
        if type_value == str or type_value == 'date':
            return js  
        try:       
            clean_js = remove_comments(js)     
            parsed = json.loads(clean_js)
        except json.JSONDecodeError as e:
            raise ValueError(f'Invalid JSON: {js}, \n Query: {str_prompt}') 
        
        if not is_type(parsed, type_value):
            raise TypeError(f'Invalid type: {type(parsed)} != {type_value}')                    
        return parsed
    return f()

def Qx(str_prompt, type_value = str):
    """returns LLM async call for a question, without formatting or extending the prompt"""
    return Q(str_prompt, type_value, format = False, extend = False)

def setup_llmrag():    
    import config #the module is loaded before config analysis    
    temperature = getattr(config, 'temperature', 0.0)
    if config.llm:
        api_key_config = None
        model = ''
        match config.llm:
            case ['host', address]:                 
                type = 'host' #provider type is openai for local llms
            case ['host', address, api_key_config, model]:                 
                type = 'host' #provider type is openai for local llms
            case [type, model, address]: ...
            case [type, model]: address = None
            case _:
                print(f'Error: Invalid llm configutation: {config.llm}') 
                return
                
        type = type.lower()
        model_kwargs={}
        reasoning = getattr(config, 'reasoning', None)
        if reasoning:
            model_kwargs['reasoning'] = {"effort": reasoning, 'enabled': True}             

        match type:
            case 'host':  
                api_key_from_config = os.environ.get(api_key_config) if api_key_config else None
                api_key = api_key_from_config if api_key_from_config else 'llm-studio'                         
                Unishare.llm_model = ChatOpenAI(
                    api_key = api_key,
                    temperature = temperature,
                    openai_api_base = address,
                    model_kwargs=model_kwargs,
                    model = model
                ) 
            case 'openai':
                
                Unishare.llm_model = ChatOpenAI(temperature = temperature, model_kwargs=model_kwargs)

            case 'xai':
                Unishare.llm_model = ChatXAI(
                    model = model,
                    temperature = temperature,
                    max_tokens = None,
                    timeout = None,
                    max_retries = 2, 
                )

            case 'groq':
                Unishare.llm_model = ChatGroq(
                    model = model,
                    temperature = temperature,
                    max_tokens = None,
                    timeout = None,
                    max_retries = 2, 
                )
            case 'google' | 'gemini':
                Unishare.llm_model = ChatGoogleGenerativeAI(
                    model = model,
                    temperature = temperature,
                    max_tokens = None,
                    timeout = None,
                    max_retries = 2,
                    safety_settings = {
                        HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_NONE
                    }
                )
            case 'mistral':
                Unishare.llm_model = ChatMistralAI(
                    model = model,
                    temperature=0,
                    max_retries=2,
                    # other params...
                )
        
        if hasattr(config, 'llm_cache'):
            Unishare.llm_cache = QueryCache(config.llm_cache)

async def get_property(name, context = '', type = str, options = None):  
    if type == str and re.search(r'date', name, re.IGNORECASE):
        type = 'date'
    limits = f', which possible options are {",".join(opt for opt in options)},' if options else ''                    
    prompt = """Output ONLY explicit value{limits} based on the context. Example: Context: Animal: Byrd. Query: Has beak: True. Context: {context}. Query: {name}:"""    
    try:
        value = await Q(prompt, type)
    except Exception as e:        
        Unishare.message_logger(e)
        return None
    return value



# Copyright © 2024 UNISI Tech. All rights reserved.
import jsonpickle, inspect, asyncio

UpdateScreen = True
Redesign = 2

def flatten(*arr):
    for a in arr:
        if isinstance(a, list | tuple):
            yield from flatten(*a)
        else:
            yield a

def index_of(lst, target):  
  try:
    return lst.index(target)
  except ValueError:
    return -1
  
async def call_anysync(handler, *params):
    return (await handler(*params)) if asyncio.iscoroutinefunction(handler) else handler(*params)

def strpath(path):
    "array path to string"
    return '@'.join(path)

def compose_handlers(*handlers):
    async def compose(obj, value):
        objs = set()        
        for handler in handlers:
            result = await call_anysync(handler, obj, value)
            if result == UpdateScreen or result == Redesign:
                return result
            if isinstance(result, list | tuple):
                for obj in flatten(result):
                    objs.add(obj)
            elif result:
                objs.add(result)
        if objs:
            return list(objs) 
    return compose
    
def equal_dicts(dict1, dict2):
    return dict1.keys() == dict2.keys() and all(dict1[key] == dict2[key] for key in dict1)

class ArgObject:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
    def __getattr__(self, _):
        """return None for unknown props"""
        return None
    @property
    def screen_type(self):
        return self.block == 'root' and self.element is None
    @property
    def voice_type(self):
        return self.block == 'voice' and self.element is None

class ReceivedMessage(ArgObject):
    def __init__(self, kwargs):
        self.__dict__.update(kwargs)
    def __str__(self):
        return f'{self.block}/{self.element}->{self.event}({self.value})'    

def toJson(obj):
    return jsonpickle.encode(obj,unpicklable = False)

def set_defaults(self, param_defaults : dict):
    for param, value in param_defaults.items():
        if not hasattr(self, param):
            setattr(self, param, value)

def pretty4(name):
    if name.startswith('_'):
        name = name[1:]
    pretty_name = name.replace('_',' ')
    return pretty_name[0].upper() + pretty_name[1:]

def is_callable(obj):
    return inspect.isfunction(obj) or inspect.ismethod(obj) or inspect.iscoroutine(obj) or callable(obj)        

def context_object(target_type):
  """
  Finds the first argument of a specific type in the current function call stack.
  """  
  frame = inspect.currentframe()
  while frame:    
    args, _, _, values = inspect.getargvalues(frame)    
    if args and isinstance(values[args[0]], target_type):
      return values[args[0]]
    # Move to the previous frame in the call stack
    frame = frame.f_back
  return None

def get_default_args(func):
    """  
    class F:
        def example_function(a, b, c=10, d='hello'):
            pass
    f = F()
    default_args = get_default_args(f.example_function)
    print(default_args)  
    """
    # Get the signature of the function
    sig = inspect.signature(func)
    # Dictionary to store arguments with their default values
    defaults = {}
    for name, param in sig.parameters.items():
        if param.default != inspect.Parameter.empty:
            defaults[name] = param.default
    return defaults

Unishare = ArgObject(context_user = lambda: None, sessions = {})

class Message:
    def __init__(self, *units, user = None, type = 'update'):        
        self.type = type        
        self.set_updates(units)
        if user:
            self.fill_paths4(user)

    def set_updates(self, units):
        self.updates = [{'data': unit} for unit in units]

    def fill_paths4(self, user):
        if hasattr(self, 'updates'):
            invisible = []
            for update in self.updates:
                data = update["data"]
                path = user.find_path(data)
                if path:
                    update['path'] = path
                else:
                    invisible.append(update)                    
                    user.log(f'Invisible element update {data.name}, type {data.type}.\n\
                    Such element not on the screen!', type = 'warning')
            for inv in invisible:
                self.updates.remove(inv)

    def contains(self, unit):
        if hasattr(self, 'updates'):
            for update in self.updates:
                if unit is update['data']:
                    return True

def TypeMessage(type, value, *data, user = None):
    message = Message(*data, user=user, type = type)    
    message.value = value
    return message    

def Warning(text, *data):
    return TypeMessage('warning', str(text), *data)

def Error(text, *data):
    return TypeMessage('error', str(text), *data)
    
def Info(text, *data):
    return TypeMessage('info', str(text), *data)

def Answer(type, message, result):
    ms = TypeMessage(type, result)
    ms.message = message
    return ms

close_message = TypeMessage('action', 'close')

def delete_unit(units, name):
    """Deletes a unit with the given name from a nested list of units.
        Returns True if the unit was found and deleted, False otherwise.
    """
    for i in range(len(units)):
        if isinstance(units[i], list | tuple):
            if delete_unit(units[i], name):
                if not units[i]: # if the sublist became empty after deletion
                    units.pop(i) # remove sublist also
                return True
        elif units[i].name == name:
            units.pop(i) 
            return True
    return False

empty_app = ArgObject(
    blocks = [],
    header = "No screens",
    icon = None,
    menu = [["You need to put at least 1 file in the 'screens' folder.",'exclamation']],
    name = "",
    order = 0,
    toolbar = [],
    type = "screen"
)







from unisi import *
from blocks.tblock import config_area, tarea

name = "Blocks"
order = 0
table = Table('Videos', 0, headers = ['Video', 'Duration',  'Links', 'Mine'], rows = [
    ['opt_sync1_3_0.mp4', '30 seconds',  '@Refer to signal1', True],
    ['opt_sync1_3_0.mp4', '37 seconds',  '@Refer to signal8', False]
])

def clear_table(_, value):
    table.rows = []    

clear_button = Button('Clear table', clear_table, icon='swipe')

selector = Select('Mode', 'All', options=['All','Based','Group'])

@handle(selector,'changed')
def selchanged(_, val):
    if val == 'Based':
        return Error('Mode can not be Based!',_)
    _.accept(val)    

def handler(*_):
    tarea.value = ''

clear_text = Button('Clear text', handler, icon='delete')

block = Block('X Block',
    [ clear_text, selector, clear_button],
    [ tarea, table], 
        icon = 'api')

def change_seletion(elem, value):
    for unit in elem.value['nodes']:        
        unit.active = False
    
    for unit in value['nodes']:
        unit.active = True
    elem.value = value

toposcreen = Net('Net', changed = change_seletion )

#graph can handle invalid edges and null nodes in the array    
graph = Graph('_Random graph', 
    nodes = [Node("Node 1"),Node("Node 2", size = 20),None, Node("Node 3", color = "green"), Node("Node 4")],
    edges = [Edge(0,1, color = "#3CA072"), Edge(1,3,'extending', size = 6),Edge(3,4, size = 2), Edge(2,4)])

def switch_graph(*_):    
    bottom_block.value[1] = toposcreen if bottom_block.value[1] == graph else graph

bottom_block = Block('Screen topology: Press Shift for multi (de)select nodes and links', 
     Button('Switch graph', switch_graph), toposcreen, closable = True)

blocks= [[block,bottom_block],config_area]

async def log(x,y):        
    for i in range(3):
        await user.send(Warning(str(i)))
    ibutton.icon = 'check'

ibutton = Button('_Ignored', lambda *_: Info('ignored!'), icon = 'delete_forever', tooltip = 'Ignore info!')

toolbar = [Button('_Save', log, icon = 'save', tooltip = 'Some info'),ibutton]

def prepare():
    if not toposcreen.topology:
        toposcreen.make_topology(blocks)



from .utils import *
from .llmrag import Q, Qx
from .units import *
from .users import User
from .server import start, handle, test, context_user, context_screen
from .tables import *
from .containers import *
from .proxy import *
from .dbunits import *
from .graphs import *
from .kdb import Database, Dbtable



from unisi import *
name = 'Single table'
order = 0

ctable = Table("Companies", id = 'Company', fields ={'name' : 'string','place': 'string','phone': 'string'})

def change_ids(button, value):
    ctable.ids = value
    ctable.calc_headers()    
    return ctable

def add_data(*_):
    rows_len = len(ctable.rows)
    data = [[f'Company {i}', f' Place {i}', f'Phone {i}']  for i in range(rows_len, rows_len + 1000)]
    ctable.rows.extend(data)        

def clear_data(*_):
    ctable.rows.clear()    

blocks = [Block('TBlock', [Switch('Show ID', False, change_ids), Button('+ data',add_data), Button('Clear data', clear_data)], ctable)]


#Hot connect to running session
from unisi import Proxy, Event

#insert session from Hello user output' #
session = '::1-0'

proxy = Proxy('localhost:8000', session = session, timeout = 7)
if proxy.event == Event.screen:
    if proxy.screen['name'] != 'Main':
        proxy.set_screen('Main')
    if proxy.event & Event.update:
        proxy.set_value('Edit string','abc')
        print('ok')
    else:
        print('error')

proxy.close()



# Copyright © 2024 UNISI Tech. All rights reserved.
from .common import Unishare
from collections import defaultdict

#storage id -> screen name -> [elem name, block name]
dbshare = defaultdict(lambda: defaultdict(lambda: []))
# db id -> [update]
dbupdates = defaultdict(lambda: [])

def at_iter(iter, times):
    for i, val in enumerate(iter):
        if i == times:
            return val
    
class Dblist:
    def __init__(self, dbtable, init_list = None, cache = None):                
        self.cache = cache
        self.limit = dbtable.limit
        if cache is not None:            
            init_list = cache[:self.limit]      
        elif init_list is None:
            raise AttributeError('init_list or cache has to be assigned!')
                              
        self.delta_list = {0 : init_list}                
        self.dbtable = dbtable                        

    def get_delta_0(self):
        return self.delta_list[0] if self.cache is None else self.cache[:self.limit]

    def __getstate__(self):        
        return dict(length = len(self),  limit = self.limit, data = self.get_delta_0())
    
    def __getattr__(self, name):
        return self.dbtable.limit if name == 'limit' else None
        
    def __iter__(self):
        "Override the default iterator to provide custom behavior."
        self._index = 0
        return self

    def __next__(self):        
        if self._index < len(self):
            value = self[self._index]
            self._index += 1
            return value
        else:
            raise StopIteration
           
    def __str__(self):
        return str(self.__getstate__())
        
    def get_delta_chunk(self, index) -> tuple[int, list]:
        """return delta list and chunk of data"""
        if index >= len(self):
            return -1, None
        delta_list = index // self.limit * self.limit
        
        if self.cache is not None:
            return delta_list, self.cache[delta_list:delta_list + self.limit]
        
        lst = self.delta_list.get(delta_list)
        if lst is not None:
            return delta_list, lst
        lst = self.dbtable.read_rows(skip = delta_list)
        self.delta_list[delta_list] = lst        
        return delta_list, lst

    def __getitem__(self, index) -> list:
        """return row from delta list or cache"""
        if self.cache is not None:
            return self.cache[index]
        delta_list, chunk = self.get_delta_chunk(index)
        if chunk:
            return chunk[index - delta_list]        

    def __setitem__(self, index, value: list):
        """update row in delta list or cache"""
        if self.cache is not None:
            self.cache[index] = value
        else:
            delta_list, chunk = self.get_delta_chunk(index)
            if chunk:
                chunk[index - delta_list] = value            
            self.dbtable.assign_row(value)
            update = dict(type = 'action', update = 'update', index = index, data = value)
            dbupdates[self.dbtable.id].append(update)            

    def clean_cache_from(self, delta_list):
        """clear dirty delta_list cache"""
        self.delta_list = {k: v for k, v in self.delta_list.items() if k < delta_list}                
        
    def __delitem__(self, index):
        delta_list, chunk = self.get_delta_chunk(index)
        if chunk:
            self.dbtable.delete_row(index)
            update = dict(type = 'action',update ='delete', index = index, exclude = True)            
            dbupdates[self.dbtable.id].append(update)
            del chunk[index - delta_list] 
            limit = self.dbtable.limit
            next_delta_list = delta_list + limit
            if len(chunk) == limit - 1: #chunk was fully filled                                
                next_list = self.delta_list.get(next_delta_list)            
                if next_list:
                    chunk.append(next_list[0])                                                                               
                self.clean_cache_from(next_delta_list)            

    def __len__(self):
        return len(self.cache) if self.cache is not None else self.dbtable.length
    
    def index2node_relation(self, cell_index):
        """calculate delta to property of node or link for persistent"""
        table_fields = self.dbtable.table_fields
        delta = cell_index - len(table_fields)       
        if delta < 0:            
            return True, at_iter(table_fields, cell_index)
        delta -= 1 #ID field
        return False, at_iter(self.dbtable.list.link[1], delta)
    
    def update_cell(self, delta, cell, value, id = None) -> dict:
        in_node, field = self.index2node_relation(cell)                    
        if in_node:
            table_id = self.dbtable.id 
            row_id =  self[delta][len(self.dbtable.table_fields)] 
        else:
            table_id = self.dbtable.list.link[2]
            row_id = id
        self.dbtable.db.update_row(table_id, row_id, {field: value}, in_node) 
        self[delta][cell] = value
        if self.cache is None: 
            update = dict(type = 'action', update = 'update', index = delta, data = self[delta])
            dbupdates[self.dbtable.id].append(update)        
            return update

    def append(self, arr):
        """append row to dblist"""
        row = self.dbtable.append_row(arr) 
        if self.cache is not None:
            self.cache.append(row)
        else:
            index = len(self) - 1                          
            delta_chunk,lst = self.get_delta_chunk(index)
            if lst is not None:
                lst.append(row)
                update = dict(type = 'action', update = 'add', index = index, data = row) 
                dbupdates[self.dbtable.id].append(update)
        return row
        
    def extend(self, rows) -> dict:
        delta_start = self.dbtable.length
        start = delta_start
        rows = self.dbtable.append_rows(rows)
        len_rows = len(rows)                
        i_rows = 0
        length = len_rows + start
        while len_rows > 0:
            delta_list = start // self.limit * self.limit
            lst = self.delta_list.get(delta_list)
            if lst is None:            
                lst = []
                self.delta_list[delta_list] = lst
                can_fill = self.limit
            else:
                can_fill = self.limit - len(lst)
            if can_fill:        
                lst.extend(rows[i_rows: i_rows + can_fill])                
                        
            i_rows += can_fill           
            start += can_fill
            len_rows -= can_fill        
        delta, data = self.get_delta_chunk(delta_start)        
        update = dict(type = 'action', update = 'updates', index = delta, data = data, length = length)
        dbupdates[self.dbtable.id].append(update)        
    
    def insert(self, index, value):
        self.append(value)        

    def remove(self, value):
        index = value[-1]
        del self[index]

    def pop(self, index = -1):
        value = self[index]
        del self[index]
        return value

    def clear(self, detach = False):
        self.dbtable.clear(detach)
        self.delta_list = {0: None}
        dbupdates[self.dbtable.id].append(dict(type = 'action', update = 'updates', length = 0))        



# Copyright © 2024 UNISI Tech. All rights reserved.
from .units import *
from .tables import Table
from .containers import Block
from collections import defaultdict
from collections.abc import Iterable

class Node:
    def __init__(self, name , color = '', size = 0, image = ''):
        if name:
            self.name = name
        if image:
            self.type = 'image'
            self.image = image
        else:
            self.type = ''
        if color:
            self.color = color
        if size:
            self.size = size        

class Edge:
    def __init__(self, source, target, name = '', color = '', size = 0, property = None):
        self.source = source
        self.target = target
        if name:
            self.name = name
        if color:
            self.color = color
        if size:
            self.size = size
        if property is not None:
            self.property = property     
    def __str__(self):
        return f"Edge({self.source}->{self.target})"   
    def __repr__(self):
        return f"Edge({self.source}->{self.target})"   

graph_default_value = {'nodes' : [], 'edges' : []}

class Graph(Unit):
    '''has to contain nodes, edges, see Readme'''
    def __init__(self, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        set_defaults(self, dict(type ='graph', value = graph_default_value, nodes = [], edges = []))

Topology = lambda: defaultdict(lambda: defaultdict(lambda: None))

def unit2image(unit):
    match unit:
        case Block():
            return 'https://img.icons8.com/fluency/48/object.png'
        case Button():
            return 'https://img.icons8.com/ios-filled/50/doorbell.png'
        case Edit() | Text():
            return 'https://img.icons8.com/fluency-systems-filled/50/123.png' if unit.type == 'number'\
                else 'https://img.icons8.com/sf-regular/48/abc.png'
        case Switch():
            return 'https://img.icons8.com/ios/50/toggle-on--v1.png'
        case TextArea():
            return 'https://img.icons8.com/color/48/align-cell-content-left.png'
        case Table():
            return 'https://img.icons8.com/color/48/day-view.png' if unit.type == 'table'\
                else 'https://img.icons8.com/ultraviolet/40/combo-chart.png'
        case Tree():
            return 'https://img.icons8.com/external-flatart-icons-outline-flatarticons/64/external-tree-nature-flatart-icons-outline-flatarticons-3.png'
        case Select():
            return 'https://img.icons8.com/cotton/64/list--v2.png'
        case Graph():
            return 'https://img.icons8.com/external-vitaliy-gorbachev-blue-vitaly-gorbachev/50/external-nodes-cryptocurrency-vitaliy-gorbachev-blue-vitaly-gorbachev.png'
        case Range():
            return 'https://img.icons8.com/ios/50/slider-control.png'
        case Unit():
            return 'https://img.icons8.com/ios-filled/50/link--v1.png'
        case _: 
            return ''

class Net(Graph):    
    """Graph of Units"""
    replace4state = dict(nodes = '_nodes', edges = '_edges', value = '_value')
    def __init__(self, name, value = graph_default_value, topology = Topology(),  **kwargs):        
        Unit.__init__(self, name, **kwargs)        
        self.type = 'graph'         
        self.value = value
        self.topology = topology       
        self._inside_converter = False
        changed_handler = getattr(self, 'changed', None)
        
        def changed_converter(_, value):                        
            self._inside_converter = True
            self._value = value
            narray = self._narray
            value = dict(nodes = [self._narray[i] for i in value['nodes']], edges = 
                [Edge(narray[self._edges[i].source], narray[self._edges[i].target]) for i in value['edges']])
            if changed_handler:
                result = changed_handler(_, value)
            else:
                result = None
                self.value = value
            self._inside_converter = False
            return result
        self.changed = changed_converter

    def specific_changed_register(self, property = None, value = None):
        """ mark serial info as invalid """
        if property:
            if property.startswith('_'):
                return False
            elif property == 'value' and self._inside_converter:
                return False
            else:
                self.delattr('_nodes')
                self.delattr('_value')                            
        return True

    def elements(self, stubs=True):
        if not hasattr(self, '_nodes'):
            self.__getstate__()
        return self.narray
    
    def __getstate__(self):
        if not hasattr(self, '_nodes'):
            nodes = []            
            narray = []
            earray = []
            for sunit, links in self.topology.items():
                sindex = index_of(narray,sunit)
                if sindex == -1:
                    sindex = len(narray)
                    narray.append(sunit)
                    nodes.append(Node(sunit.name, image = unit2image(sunit), 
                            color = 'white', size = 15))
                for dunit  in links:
                    dindex = index_of(narray,dunit)
                    if dindex == -1:
                        dindex = len(narray)
                        narray.append(dunit)
                        nodes.append(Node(dunit.name, image = unit2image(dunit), 
                            color = 'white', size = 15))
                    earray.append(Edge(sindex, dindex))
            self._nodes = nodes
            self._edges = earray
            self._narray = narray            
        
        if not hasattr(self, '_value'):
            self._value = dict(nodes = [index_of(self._narray,unit) for unit in self.value['nodes']], 
            edges = [Edge(index_of(self._narray, e.source), index_of(self._narray, e.target)) for e in self.value['edges']])            
        return {name: getattr(self,Net.replace4state.get(name, name)) for name in ['nodes', 'edges', *self.__dict__.keys()] 
                                                                        if name != 'topology' and name[0] != '_'}    

    def make_topology(self, unit: Unit | Iterable):
        topo = Topology()
        def dive(unit):
            match unit:
                case Iterable():
                    node = Unit('Union', type = 'union')            
                    for obj in unit:
                        if obj:
                            topo[node][dive(obj)] = {}                
                    return node
                case Block():
                    for obj in unit.value:
                        if obj:
                            topo[unit][dive(obj)] = {}                
                case _: ...
            return unit
        dive(unit)
        self.topology = topo
        self.specific_changed_register()



# Copyright © 2024 UNISI Tech. All rights reserved.
from difflib import SequenceMatcher
from word2number import w2n 
from .users import *
from .units import *
from .containers import Block

def find_most_similar_sequence(input_string, string_list):    
    best_match = ""
    highest_ratio = 0
    for string in string_list:
        matcher = SequenceMatcher(None, input_string.lower(), string.lower())
        ratio = matcher.ratio()
        if ratio > highest_ratio:
            highest_ratio = ratio
            best_match = string
    return best_match, highest_ratio

def word_to_number(word: str):
    sn = word.replace(',', '')
    try: 
        return float(sn) 
    except:
        try: 
            return w2n.word_to_num(sn) 
        except:
            return None
    
command_synonyms = dict( #-> words    
    value = ['is', 'equals'],
    root = ['select', 'choose','set'],
    backspace = ['back'],    
    enter = ['push', 'execute','run'],
    clean = ['empty','erase'],
    screen = ['menu'], 
    push = ['execute','run'],         
    reset = ['cancel'],
    ok = ['okay']
)

root_commands = ['root', 'screen', 'stop', 'reset', 'ok']   
ext_root_commands = root_commands[:]

modes = dict( #-> actions    
    text = ['text', 'left', 'right', 'up', 'down','backspace','delete', 'space', 'tab', 'enter', 'undo','clean'],
    number = ['number', 'backspace','delete', 'undo','clean'],    
    graph = ['node', 'edge'],
    table = ['page','row', 'column', 'left', 'right', 'up', 'down','backspace','delete'], 
    command = ['push']   
)
word2command = {}
for command, syns in command_synonyms.items():
    for syn in syns:
        word2command[syn] = command          
    if command in root_commands:
        ext_root_commands.extend(syns)

word2command.update({command: command for command in root_commands})
for mode, commands in modes.items():
    word2command.update({c:c for c in commands})
    

class VoiceCom:
    def __init__(self, user):
        self.block = self.assist_block(user)    
        self.user = user    
        self.previous_unit_value_x = None
        self.unit = None
        self.cached_commands = {}
        self.set_screen(user.screen)

    async def keyboard_input(self, _, value):
        return await self.process_string(value)

    def assist_block(self, user) -> Block:
        self.input = Edit('Recognized words', '', update = self.keyboard_input)
        self.message = Edit('System message', '', edit = False)
        self.context_list = Select('Elements', None, self.select_elem, type = 'list', width = 250)
        self.command_list = Select('Commands', None, self.select_command, type = 'list', width = 250)
        
        block = Block("Mate:", 
            self.message,
            self.input,            
            self.context_list,
            self.command_list,
            closable = True, icon = 'mic'                                
        )      
        block.set_reactivity(user)  
        return block
    
    def set_screen(self, screen):        
        self.calc_interactive_units()
        self.screen = screen
        self.reset()

    def select_elem(self, elem, value):
        elem.value = value
        if value:
            if self.mode == 'screen':
                self.user.set_screen(value)
            self.activate_unit(self.name2unit.get(value, None))    

    async def select_command(self, _, value):
        _.value = None        
        return await self.process_word(value)        

    @property
    def  context_options(self):
        return self.context_list.options

    @context_options.setter
    def context_options(self, names):
        self.context_list.options = names       

    @property
    def commands(self):
        return self.command_list.options
        
    @commands.setter
    def commands(self, commands):
        self.command_list.options = commands

    @property
    def context(self):
        return self.context_list.value
    
    @context.setter
    def context(self, context):
        self.context_list.value = context        
                
    def calc_interactive_units(self):
        interactive_names = []
        name2unit = {}        
        self.sreen_name = self.user.screen.name        
        for block in flatten(self.user.screen.blocks):                        
            for elem in flatten(block.value):
                if getattr(elem, 'edit', True):
                    pretty_name = pretty4(elem.name)
                    name2unit[pretty_name] = elem
                    interactive_names.append(pretty_name)  
        interactive_names.sort()                  
        self.unit_names = interactive_names
        self.name2unit = name2unit
        
    def set_mode(self, mode):
        self.context = None
        self.mode = mode
        self.buffer = []
        self.previous_unit_value_x = None
        if not (commands := self.cached_commands.get(mode, None)):            
            syn_commands = []
            commands = modes.get(mode, []) + root_commands
            for command in commands:
                if command in command_synonyms:
                    syn_commands.extend(command_synonyms[command])        
            commands.extend(syn_commands)        
            commands.sort()
            self.cached_commands[mode] = commands
        self.commands = commands
        self.input.value = mode
        self.message.value = 'Continue..'
        match self.mode:
            case 'switch' | 'check':
                self.context_options = ['true', 'false','yes', 'no', 'on', 'off']
            case 'select' | 'list' | 'radio':
                self.context_options = self.unit.options
            case 'tree':
                self.context_options = list(self.unit.options)  
            case 'screen':
                self.context_options = [getattr(s, 'name')for s in self.user.screens 
                    if hasattr(s, 'name') and s.name != self.user.screen.name]
                self.message.value = 'Select a screen'          
            case _:
                self.context_list.options = []
        self.context = None        
    
    def activate_unit(self, unit: Unit | None):
        if self.unit:
            self.unit.active = False
            self.unit.focus = False
        self.unit = unit
        self.message.value =  'Select a command'
        if unit:
            match unit.type:
                case 'string':
                    mode = 'text'
                case 'range':
                    mode = 'number'
                case _: 
                    mode = unit.type        
            unit.active = True
            unit.focus = True            
            self.set_mode(mode)
            if unit.type == 'text' or unit.type == 'number':
                self.previous_unit_value_x = unit.value, unit.x
        else:
             self.commands = ext_root_commands                    

    def start(self):
        if self.screen.blocks[-1] != self.block:
            self.screen.blocks.append(self.block)    
        self.reset()        
    def stop(self):
        if self.screen.blocks[-1] == self.block:
            self.screen.blocks.remove(self.block)            

    async def process_string(self, string: str) -> any:  
        screen_changed = None
        for word in string.split(' '):
            if word:
                result = await self.process_word(word)
                if not screen_changed:
                    screen_changed = result
        return screen_changed 

    async def process_word(self, word: str) -> any:  
        self.input.value = word
        self.message.value = ''
        if word:      
            command = word2command.get(word, None)
            match self.mode:
                case 'number'|'text':
                #double repeat command word cause execution
                    if self.mode == 'number':
                        num = word_to_number(word)
                        if command:
                            return await self.run_command(command)
                        elif num is not None:
                            self.previous_unit_value_x = self.unit.value, self.unit.x                            
                            self.unit.value = num                                                    
                        else:
                            self.message.value = 'Not a number'
                    elif self.mode == 'text':
                        if self.buffer and self.buffer[-1] == word and command:
                            self.buffer.pop()
                            if self.previous_unit_value_x:
                                self.unit.value, self.unit.x = self.previous_unit_value_x                            
                                self.previous_unit_value_x = None
                            return await self.run_command(command)
                        else:
                            self.previous_unit_value_x = self.unit.value, self.unit.x
                            self.buffer = [word]
                            value = self.unit.value
                            if self.unit.x == -1:
                                if value:
                                    self.unit.value += ' ' + word
                                else:
                                    self.unit.value = word                                                    
                                self.unit.x = len(self.unit.value)
                            else:                 
                                word += ' '           
                                self.unit.value = value[:self.unit.x] + word + value[self.unit.x:]
                                self.unit.x += len(word)            
                case 'switch' | 'check' | 'select' | 'list' |'radio' | 'tree':
                    if command:
                        return await self.run_command(command)
                    if self.context:                        
                        self.message.value = ''
                        self.unit.value = self.context in ('true', 'yes', 'on') if self.mode == 'switch' else self.context
                    else:
                        choice, similarity = self.buffer_suits_name(word)                    
                        if similarity >= 0.8:
                            self.unit.value = choice in ('true', 'yes', 'on') if self.mode == 'switch' else choice
                            self.message.value = ''                                
                        elif choice:
                            self.context = choice
                            self.message.value = '"Ok" to confirm'
                        else:
                            self.commands = []
                            self.message.value = 'Continue..'
                            self.buffer = []
                            self.input.value = ''                    
                case 'root': 
                    if command == 'ok' and self.context:
                        self.activate_unit(self.name2unit[self.context])
                    elif command:
                        return await self.run_command(command)
                    else:
                        unit_name, similarity = self.buffer_suits_name(word)                    
                        if similarity >= 0.8:
                            self.activate_unit(self.name2unit[unit_name])
                        elif unit_name:
                            self.context = unit_name
                            self.message.value = '"Ok" to confirm'
                        else:
                            self.commands = []
                            self.message.value = 'Continue..'
                            self.input.value = ' '.join(self.buffer)                      
                case 'screen':
                    if command == 'ok' and self.context:
                        self.user.set_screen(self.context)
                    else:
                        screen_name, similarity = self.buffer_suits_name(word)
                        if similarity > 0.9:
                            self.user.set_screen(screen_name)
                        else:
                            self.context = screen_name
                            self.message.value = '"Ok" to confirm'             
                case _:
                    if command: 
                        return await self.run_command(command)  
                    else:       
                        self.message.value = 'Unknown command.'

    async def run_command(self, command: str):  
        self.message.value = ''                            
        match command:
            case 'root':            
                self.reset()                
            case 'screen':                
                self.set_mode('screen')                                
            case 'reset':
                self.reset()
            case 'stop':
                self.stop()
            case _:
                if self.unit:
                    return await self.context_command(command)
                else:
                    self.message.value = 'Command is out of context.'

    def reset(self):
        self.buffer = []
        self.mode = 'root'
        if dialog := self.user.active_dialog:
            commands = ext_root_commands + ['close']
            commands.sort
            self.commands = commands
            options = [u.name for u in flatten(dialog.value)]
            options.sort()
            self.context_options = options            
        else:
            self.commands = ext_root_commands
            self.context_options = self.unit_names
        if self.unit:
            self.unit.active = False
            self.unit.focus = False
            self.unit = None
        self.input.value = ''
        self.message.value = 'Select a command or element'
        self.context = None        

    def buffer_suits_name(self, word: str):
        self.buffer.append(word)
        name = ' '.join(self.buffer)         
        return find_most_similar_sequence(name, self.context_options)
                    
    async def context_command(self, command: str):
        u = self.unit                   
        match self.mode:                        
            case 'text':
                match command:
                    case 'left' :
                        if u.x > 0:
                            u.x -= 1
                    case 'right' :
                        if u.x < len(u.value) - 1:
                            u.x += 1                    
                    case 'backspace' :
                        if u.x > 0:
                            u.value = u.value[:u.x - 1] + u.value[u.x:]
                            u.x -= 1
                    case 'delete' :
                        if u.x < len(u.value) - 1:
                            u.value = u.value[:u.x] + u.value[u.x+1:]
                    case 'space' :
                        if hasattr(u, 'x'):
                            u.value = u.value[:u.x] + ' ' + u.value[u.x:]
                            u.x += 1
                    case 'undo':
                        if self.previous_unit_value_x:
                            u.value, u.x = self.previous_unit_value_x 
                    case 'clean':
                        u.value = ''
                    case _ :
                        self.message.value = 'Command is ouside context'
            case 'number':
                svalue = str(u.value)
                match command:
                    case 'left' :
                        if u.x > 0:
                            u.x -= 1
                    case 'right' :
                        if u.x < len(svalue) - 1:
                            u.x += 1
                    case 'backspace' :
                        if u.x > 0:
                            u.value = float(svalue[:u.x - 1] + svalue[u.x:])
                            u.x -= 1
                    case 'delete' :
                        if u.x < len(svalue) - 1:
                            u.value = float(svalue[:u.x] + svalue[u.x+1:])
                    case 'clean':
                        u.value = None
                    case _ : 
                        self.message.value = 'Command is ouside context'
                                                
            case 'command':
                if command == 'ok' or command == 'push':
                    handler = self.unit.changed
                    if handler:
                        return await call_anysync(handler, self.unit, None)                  
                self.message.value = 'Command is ouside context'                                    


from torch import Block
from unisi import *
import random, copy, time
from data import long_function

def append(_,val):
    ''' append has to return new row'''
    id, search = val
    new = [search, '', '']
    _.rows.append(new)
    return new

def updated(_, value):
    _.value = value  
    return Info(f'{_.name} is updated to {value}!')        

def complete(_, value):    
    value, pos = value          
    return ['aaa', 'bbb', 'ccc']   if pos[1] == 0 else [] 

def changed(_, value):
    _.value = value  
    return Warning(f'{_.name} changed to {value}!')

def table_modify(_, value):
    value, pos = value    
    return Error(f'{_.name} {pos} is not modified to value {value}!', _)

def table_update(_, value):    
    accept_cell_value(_, value)
    value, pos = value        
    return Info(f'{_.name} {pos} is updated to value {value}!')

ticks = Edit('How many ticks', 50)

async def serious_function(ticks):
    await user.progress('Run process')
    for i in range(ticks):        
        time.sleep(0.04)
        await user.progress(f'{i} tick')
    return 5


async def dialog_callback(_,value):    
    if value == 'Ok':
        user = context_user()        
        #bug in python-windows in call multiprocessing.Manager()
        #res = await user.run_process(long_function, ticks.value, None, progress_callback = user.progress)      
        res = await serious_function(50)
        txt.value = f'Process result is {res}'
        return txt      
        
def call_dialog(*_):
    return Dialog('Start a long process?', dialog_callback, ticks)

table = Table('Audios', 1, changed, type = 'chart', headers = ['Audio', 'Duration,sec', 'Stars'], multimode = True,
    rows =  [[f'sync{i}.mp3', round(random.random() * 15000) / 100, random.randint(1,50)] for i in range(100)],
    append = append, complete = complete, update = table_update, view = 'i-1,2')

tarea = TextArea('Area', "Alpha Centauri A, also known as Rigil Kentaurus, is the principal member, or primary,\
 of the binary system. It is a solar-like main-sequence star with a similar yellowish colour, whose stellar \
classification is spectral type G2-V; it is about 10% more massive than the Sun, with a radius about 22% larger.",
    update = updated)

ld = { 
    'Animals' : None,
    'Brushtail Possum' : 'Animals',
    'Genet' : 'Animals',
    'Silky Anteater' : 'Animals',
    'Greater Glider' : 'Animals',
    'Tarsier' : 'Animals',
    'Kinkajou' : 'Animals',
    'Tree Kangaroo' : 'Animals',
    'Sunda Flying Lemur' : 'Animals',
    'Green Tree Python' : 'Animals',
    'Fruit Bat' : 'Animals',
    'Tree Porcupines' : 'Animals',
    'Small Tarsier' : 'Tarsier',
    'Very small Tarsier': 'Small Tarsier'
}

def tree_changed(_, item): 
    _.value = item 
    return Info(f"{item} selected in the tree.")

tree = Tree('_Inheritance','Animals', tree_changed, options = ld)

txt = Text('Text about cats')

simple_enter = Edit('_Simple Enter update', 'cherokke', update = updated)

list_complete = ['value 21', 'Value 21', 'sum 289', 'absolute 3']

eblock = Block('Shared block',                        
        [Button('Dialog for a process', call_dialog),Range('Scaling', 0, changed, options=[0.0,1.0,0.1])],        
        txt,
        Edit('Number only', 2.5),
        Block('Embedded block', Button('Inner button', call_dialog),
            Edit('Edit string', 'xyz'),
            Edit('Complete enter update field', '', changed, complete = smart_complete(list_complete), update = updated))            
)

def add_tree_elem(_, val):
    txt = simple_enter.value
    if not txt:
        return Info('Enter text first to the field!')
    if txt in ld:
        return Warning('Cannot add doubles!')    
    ld[txt] = tree.value if tree.value else None
    tree.value = txt    

treeblock = Block('Tree block',[simple_enter, Button('_Add to tree', add_tree_elem, icon='add_circle')], tree, icon = 'account_tree')

tableblock = Block('Table block', [], table, icon = 'insights')

config_area = [eblock, [treeblock, tableblock]]




# Copyright © 2024 UNISI Tech. All rights reserved.
from .autotest import config

if not config.hot_reload:
    active_reloader = False
else:
    active_reloader = True
    import os, sys, traceback
    from watchdog.observers import Observer
    from watchdog.events import PatternMatchingEventHandler
    from .users import User, Redesign
    from .utils import divpath, app_dir
    from .autotest import check_module
    import re, collections

    #for removing message duplicates        
    file_content = collections.defaultdict(str)
    
    busy = False      
    cwd = os.getcwd()  

    def free():
        global busy
        if request_file:
            reload(request_file)
        else:
            busy = False

    def reload(sname, changed_dependency = False):
        user = User.last_user
        if user:
            file = open(f'screens{divpath}{sname}', "r") 
            content = file.read()
            if not changed_dependency and file_content[sname] == content:
                return
            file_content[sname] = content
            
            global busy, request_file
            busy = True
            request_file = None            

            try:
                module = user.load_screen(sname)
                errors = check_module(module)                                
                if errors:
                    print('\n'.join(errors))
                    busy = False
                    return 
                print('Reloaded.') 
            except:
                traceback.print_exc()        
                busy = False                
                return

            for i, s in enumerate(user.screens):
                if s.__file__ == module.__file__:
                    same = user.screen_module.__file__ == module.__file__
                    user.screens[i] = module
                    if same:
                        user.set_screen(module.screen.name)            
                    break
            else:
                user.screens.append(module)
                if len(user.screens) == 1:
                    user.set_screen(module.name)                    

            user.screens.sort(key=lambda s: s.screen.order)           
            menu = [[getattr(s, 'name', ''),getattr(s,'icon', None)] for s in user.screens]               
            for s in user.screens:
                s.screen.menu = menu
            user.set_clean() 
            if hasattr(user,'send'):
                user.sync_send(Redesign)
            free()  
            return module  

    class ScreenEventHandler(PatternMatchingEventHandler):    
        def on_modified(self, event):
            if not event.is_directory and User.last_user:                            
                short_path = event.src_path[len(cwd) + 1:]
                arr = short_path.split(divpath) 
                name = arr[-1]
                dir = arr[0] if len(arr) > 1 else '' 
                                
                if name.endswith('.py'):
                    user = User.last_user
                    
                    changed_dependency = False
                    if user.screen_module and dir not in ['screens','blocks']: 
                        changed_dependency = True
                        #analyze if dependency exist
                        file = open(user.screen_module.__file__, "r") 
                        arr[-1] = arr[-1][:-3]
                        module_name = '.'.join(arr) 
                        module_pattern = '\.'.join(arr)                        
                        
                        if re.search(f"((import|from)[ \t]*{module_pattern}[ \t\n]*)",file.read()):
                            if module_name in sys.modules:
                                del sys.modules[module_name]                            
                            short_path = user.screen_module.__file__
                            if short_path.startswith(app_dir):
                                short_path = short_path[len(app_dir) + 1:]
                            dir, name = short_path.split(divpath)                            

                    if dir in ['screens','blocks']:                             
                        if busy:
                            global request_file            
                            request_file = short_path 
                        else:                    
                            fresh_module = reload(name, changed_dependency) if dir == 'screens' else None                                            
                            module = user.screen_module
                            if module:
                                current = module.__file__
                                if not fresh_module or current != fresh_module.__file__:
                                    reload(current.split(divpath)[-1], changed_dependency) 
                                                    
        def on_deleted(self, event):            
            if not event.is_directory and User.last_user:
                user = User.last_user            
                arr = event.src_path.split(divpath) 
                name = arr[-1]
                dir = arr[-2]  
                if name.endswith('.py') and dir == 'screens':
                    delfile = f'{dir}{divpath}{name}'
                    for i, s in enumerate(user.screens):
                        if s.__file__ == event.src_path:
                            user.screens.remove(s)
                            if user.screen_module is s:
                                if user.screens:                                                                        
                                    fname = user.screens[0].__file__.split(divpath)[-1]
                                    module = reload(fname)
                                    user.set_screen(module.name)
                                    user.update_menu()   
                                    if hasattr(user,'send'):                             
                                        user.sync_send(Redesign)      
                                else:   
                                    if hasattr(user,'send'):                                                   
                                        user.sync_send(empty_app)                                                        
                            else:
                                reload(user.screen_module.__file__.split(divpath)[-1])
                                user.update_menu()
                                if hasattr(user,'send'):
                                    user.sync_send(Redesign)                                                        
                            break
    
    event_handler = ScreenEventHandler()
    observer = Observer()
    path = os.getcwd()
    observer.schedule(event_handler, path, recursive = True)
    observer.start()
    


import re
from abc import ABC


class Ignore(ABC):

    @classmethod
    def transform(cls, obj, rules):
        t = type(rules)
        if obj:
            if t is dict:
                return cls._apply_dictable_rule(obj, rules)
            if t is list:
                return cls._apply_listable_rule(obj, rules)
        return obj

    @classmethod
    def _apply_dictable_rule(cls, obj, rules):
        for key in rules:
            rule = rules[key]
            if cls._is_special_key(key):
                obj = cls._apply_special_rule(key, obj, rule)
            elif cls._is_regex_rule(rule):
                obj = cls._apply_regex_rule(key, obj, rule)
            elif type(rule) is str:
                obj = cls._apply_stringable_rule(key, obj, rule)
            elif key in obj:
                obj[key] = cls.transform(obj[key], rule)
        return obj

    @classmethod
    def _apply_listable_rule(cls, obj, rules):
        for key in rules:
            if type(key) is dict:
                for index, y in enumerate(obj):
                    obj[index] = cls.transform(obj[index], key)
            elif key in obj:
                del obj[key]
        return obj

    @classmethod
    def _apply_stringable_rule(cls, key, obj, rule):
        if rule == '*':
            if key in obj:
                del obj[key]
        return obj

    @classmethod
    def _is_regex_rule(cls, rule):
        return type(rule) is dict and '_re' in rule

    @classmethod
    def _apply_regex_rule(cls, key, obj, rule):
        regex = rule['_re']
        if key in obj and re.match(regex, obj[key]):
            del obj[key]
        return obj

    @classmethod
    def _is_special_key(cls, key):
        return key.startswith('_')

    @classmethod
    def _apply_special_rule(cls, key, obj, rule):
        if key == '_values':
            return cls._ignore_values(obj, rule)
        if key == '_list':
            return cls._ignore_list_items(obj, rule)
        if key == '_range':
            return cls._ignore_range(obj, rule)
        return obj

    @classmethod
    def _ignore_list_items(cls, obj, rule):
        return [cls.transform(x, rule) for x in obj]

    @classmethod
    def _ignore_values(cls, obj, black_list):
        t = type(obj)
        if t is list:
            return [x for x in obj if x not in black_list]
        if t is dict:
            return {k: obj[k] for k in obj if k not in black_list}
        return obj

    @classmethod
    def _ignore_range(cls, obj, rule):
        t = type(obj)
        if t is int or t is float:
            return rule[0] <= obj and obj <= rule[1]
        return obj



# Copyright © 2024 UNISI Tech. All rights reserved.
from websocket import create_connection
from enum import IntFlag
import json, requests, os
from .common import *

class Event(IntFlag):
    none = 0
    update = 1
    invalid = 2
    message = 4
    update_message = 5
    progress = 8
    update_progress = 9
    unknown = 16
    unknown_update = 17
    dialog = 32
    screen = 65
    complete = 128
    append = 256    

ws_header = 'ws://'
wss_header = 'wss://'
ws_path = 'ws'

message_types = ['error','warning','info']

class Proxy:
    """UNISI proxy"""
    def __init__(self, host_port, timeout = 7, ssl = False, session = ''):
        addr_port = f'{wss_header if ssl else ws_header}{host_port}'
        addr_port = f'{addr_port}{"" if addr_port.endswith("/") else "/"}{ws_path}'
        self.host_port = f'{"https" if ssl else "http"}://{host_port}'
        if session:
            addr_port = f'{addr_port}?{session}'
        self.conn = create_connection(addr_port, timeout = timeout)
        self.screen = None       
        self.screens = {}
        self.dialog = None
        self.event = None
        self.request(None)

    def close(self):
        self.conn.close()
    
    @property
    def screen_menu(self):
        return [name_icon[0] for name_icon in self.screen['menu']] if self.screen else []
    
    @property
    def commands(self):
        """return command objects"""
        return self.elements(types=['command'])        
            
    def element(self, name, block_name = None):
        """return the element only if 1 element has such name"""
        result = None
        name2block = self.screen['name2block']
        for block in [name2block[block_name]] if block_name else name2block.values(): 
            for el in flatten(block['value']):
                if el['name'] == name:
                    if not result:
                        result = el
                    else:
                        return None
        return result
    
    def elements(self, block = None, types = None):
        """get elements with filtering types and blocks"""
        if block:
            return [el for el in flatten(block['value']) if not types or el['type'] in types]
        answer = []
        for block in self.screen['name2block'].values(): 
            answer.extend([el for el in flatten(block['value']) if not types or el['type'] in types])
        return answer
    
    def block_name(self, element):
        is_name = isinstance(element, str)
        for block in self.screen['name2block'].values():
            for el in flatten(block['value']):
                if el['name'] == element if is_name else el == element:
                    return block['name']
                
    def upload(self, fpath):
        """upload file to the server and get its server path"""
        file = open(fpath, "rb")        
        response = requests.post(self.host_port, files = {os.path.basename(fpath): file})
        return getattr(response, 'text', '')
    
    def command(self, command, value = None):
        return self.interact(self.make_message(command, value))        
        
    def command_upload(self, command, fpath):
        """upload file to the server and call command"""    
        spath = os.path.abspath(fpath) if 'localhost' in self.host_port else self.upload(fpath)
        return self.command(command, spath) if spath else Event.invalid                    
    
    def make_message(self, element, value = None, event = 'changed'):
        if isinstance(element, str):
            element = self.element(element)            
        if event != 'changed' and event not in element:
            return None        
        return ArgObject(block = self.block_name(element), element = element['name'], 
            event = event, value = value)
    
    def interact(self, message, progress_callback = None):
        """progress_callback is def (proxy:Proxy)"""        
        while self.request(message) & Event.progress:
            if progress_callback:
                progress_callback(self)
            message = None
        return self.event
    
    def request(self, message):
        """send message or message list, get responce, return the responce type"""
        if message:
            self.conn.send(toJson(message))
        responce = self.conn.recv()
        message = json.loads(responce) 
        return self.process(message) 
    
    def set_value(self, element, new_value):
        if isinstance(element, str):
            element = self.element(element)            
        element['value'] = new_value
        ms = self.make_message(element, new_value)
        return self.interact(ms) if ms else Event.invalid
    
    def set_screen(self, name):
        screen = self.screens.get(name)
        if not screen:
            if name in self.screen_menu:
                mtype = self.request(ArgObject(block = 'root', element = None, value = name))
                return mtype == Event.screen 
            else:
                return False
        return True 
    
    @property
    def dialog_commands(self):
        return self.dialog['commands'] if self.dialog else []
    
    def dialog_responce(self, command: str | None):
        if not self.dialog:
            self.event = Event.invalid
            return  self.event
        return self.interact(ArgObject(block = self.dialog['name'], value = command))
       
    def process(self, message):        
        self.message = message   
        if not message:
            self.event = Event.none 
            self.mtype = None
        else:    
            mtype = message.get('type')        
            self.mtype = mtype
            if mtype == 'screen':
                self.screen = message
                self.screens[self.screen['name']] = message
                name2block = {block['name']: block for block in flatten(message['blocks'])}            
                name2block['toolbar'] = {'name': 'toolbar', 'value': message['toolbar']}
                message['name2block'] = name2block
                self.event = Event.screen
            elif mtype == 'dialog':
                self.dialog = message
                self.event = Event.dialog
            elif mtype == 'complete':
                return Event.complete
            elif mtype == 'append':
                self.event = Event.append
            elif mtype == 'update':
                self.update(message)
                self.event = Event.update
            else:
                updates = message.get('updates')
                if updates:
                    self.update(message)                                        
                if type in message_types:
                    self.event = Event.update_message if updates else Event.message 
                if type == 'progress':
                    self.event = Event.update_progress if updates else Event.progress
                else:
                    self.event = Event.unknown_update if updates else Event.unknown
        return self.event
        
    def update(self, message):
        """update screen from the message"""
        result = Event.update
        updates = message.updates
        for update in updates:
            path = update['path']
            name2block = self.screen['name2block']
            if len(path) == 1: #block
                name2block[block] = update['data']
            else:
                block, element = path
                for el in flatten(name2block[block]['value']):
                    if el['name'] == element:
                       el.__dict__ =  update['data'].__dict__ 
                       break
                else:
                    result = Event.unknown_update
        return result

    



#Interact with https://github.com/unisi-tech/vision
from unisi import Proxy, Event

proxy = Proxy('localhost:8000')

#image for analysis
image_file = '/home/george/Projects/save/animals/badger/0cf04d0dab.jpg'

if proxy.set_screen("Image analysis"):    
    #optional: turn off search images for perfomance, we only need to classify the image
    #for that find Switch 'Search' and set it to False    
    proxy.set_value('Search', False)
    
    if proxy.command_upload('Load an image', image_file) & Event.update:
        table = proxy.element('Image classification')        
        print('  Answer:')
        for row in table['rows']:
            print(row)

proxy.close()



from .compare import NO_DIFF, Compare
from .errors import (
    KeyNotExist,
    LengthsNotEqual,
    TypesNotEqual,
    UnexpectedKey,
    ValueNotFound,
    ValuesNotEqual,
)

__all__ = (
    "Compare",
    "NO_DIFF",

    "ValuesNotEqual",
    "TypesNotEqual",
    "KeyNotExist",
    "ValueNotFound",
    "LengthsNotEqual",
    "UnexpectedKey",
)



# Copyright © 2024 UNISI Tech. All rights reserved.
import config, os, json, asyncio
from .utils import *
from .units import * 
from .containers import Block, Dialog
from .users import User
from .common import *
from .jsoncomparison import Compare, NO_DIFF

def obj2json(obj):
    return json.loads(toJson(obj))

if not os.path.exists(config.upload_dir):
    os.makedirs(config.upload_dir)

comparator = Compare(rules = {'toolbar': '*'}).check

class Recorder:
    def __init__(self):
        self.start(None)
        
    def accept(self, msg, response):  
        if not self.ignored_1message: 
           self.ignored_1message = True   
        else:           
            self.record_buffer.append(f"{toJson(msg)},\
              \n{'null' if response is None else toJson(response)}\n")
        
    def stop_recording(self, _, x):    
        button.spinner = None
        button.changed = button_clicked
        button.tooltip = 'Create autotest'
        full = len(self.record_buffer) > 1
        if full:             
            with open(self.record_file, mode='w') as file:    
                content = ',\n'.join(self.record_buffer)
                file.write(f"[\n{content}]")
        test_name = self.record_file
        self.record_file = None
        
        return Info(f'Test {test_name} is created.', button) if full else\
            Warning('Nothing to save!',button)

    def start(self,fname):
        self.record_file = fname        
        self.record_buffer = []
        if fname:
            self.ignored_1message = True
            module = User.last_user.screen_module
            self.accept(ArgObject(block = 'root', element = None,
                event = 'changed', value = module.name), module.screen)
            self.ignored_1message = False

recorder = Recorder()

def test(filename, user):
    filepath = f'{testdir}{divpath}{filename}'
    file = open(filepath, "r") 
    data = json.loads(file.read())    
    error = False
    for i in range(0, len(data), 2):
        message = data[i]
        expected = data[i + 1]
            
        result = asyncio.run(user.result4message(ReceivedMessage(message)))
        responce = user.prepare_result(result)
        jresponce = obj2json(responce)
        
        diff = comparator(expected, jresponce)
        if diff != NO_DIFF:
            print(f"\nTest {filename} is failed on message {message}:")            
            for key, obj in diff.items():                                             
                if key != '#name':
                    while True:
                        err = obj.get('_message')
                        if err:
                            print(f"  {err} \n")
                            break
                        else: 
                            content = obj.get('_content')
                            if content and len(obj) == 1:
                                obj = content
                            else:
                                for key, subobj in obj.items():
                                    if key != '_content': 
                                        if isinstance(key, str):  
                                            name = obj.get('#name', '')                                                                             
                                            if name:
                                                key = f'  {name}: {key}'
                                            print(f"  {key}")
                                        obj = subobj
                                        break                                                                
            error = True
    return not error

test_name = Edit('Name test file', '', focus = True)
rewrite = Switch('Overwrite existing', False, type = 'check')

def button_clicked(_,__):
    if not os.path.exists(testdir):
        os.makedirs(testdir)        
    test_name.value = User.last_user.screen.name    
    return Dialog('Create autotest..', ask_create_test, test_name, rewrite)

def create_test(fname):
    fname = f'{testdir}{divpath}{fname}'    
    if os.path.exists(fname) and not rewrite.value:
        return Warning(f'Test file {fname} already exists!')              
    
    button.spinner = True   
    button.tooltip = 'Stop test recording'
    button.changed = recorder.stop_recording
    recorder.start(fname)
    
    return Info('Test is recording.. press the same button to stop',button)     

def ask_create_test(_, bname):
    if bname == 'Ok':            
        return create_test(test_name.value) if test_name.value else\
            Warning('Test file name is not defined!')
    
button = Button('_Add test', button_clicked, right = True,
        icon='format_list_bulleted_add', tooltip='Create autotest')

if config.autotest:            
    User.toolbar.append(button)

def check_block(block, hash_elements):
    errors = []
    child_names = set()   
    
    if not hasattr(block, 'name') or not block.name:            
        errors.append(f"The block with {[str(type(gui)).split('.')[-1] for gui in flatten(block.value)]} does not contain name!")
        block.name = 'Unknown'          
    if not isinstance(block.name, str):
        errors.append(f"The block name `{block.name}` is not a string!")
    for child in flatten(block.value):  
        hash_element = hash(child)
        if hash_element in hash_elements:
            errors.append(f'The block {block.name} contains already used "{child.name}" in block "{hash_elements[hash_element]}"!')
        else: 
            hash_elements[hash_element] = block.name         
            if not isinstance(child, Unit) or not child:
                errors.append(f'The block {block.name} contains invalid element {child} instead of Unit+ object!') 
            elif isinstance(child, Block):
                errs = check_block(child, hash_elements)
                if errs:
                    errors += errs
            if child.name in child_names and child.type != 'line':                        
                errors.append(f'The block {block.name} contains a duplicated element name "{child.name}"!')
            elif child.type == 'chart' and not hasattr(child, 'view') and not hasattr(child, 'option'):
                errors.append(f'The block {block.name} contains a chart type "{child.name}", but "view" or "option" does not defined!')
            else:
                child_names.add(child.name)                
    return errors

def check_module(module):
    screen = module.screen
    errors =  []        
    block_names = set()      
    hash_elements = {}  
    if not hasattr(screen, 'name') or not screen.name:            
        errors.append(f"Screen file {module.__file__} does not contain name!")
        screen.name = 'Unknown'
    elif not isinstance(screen.name, str):
        errors.append(f"name' variable in screen file {module.__file__} {screen.name} is not a string!")
    if not isinstance(screen.blocks, list):
        errors.append(f"Screen file {module.__file__} 'blocks' is not a list!")
    else:
        toolbar = 'toolbar'
        block_names.add(toolbar)
        if screen.toolbar:
            errors += check_block(Block(toolbar, *screen.toolbar), hash_elements)
        for bl in flatten(screen.blocks):            
            if not isinstance(bl, Block):
                errors.append(f'The screen contains invalid element {bl} instead of Block object!')                                                    
            elif bl.name in block_names:
                errors.append(f'The screen contains a duplicated block name {bl.name}!')    
            else:            
                block_names.add(bl.name)
            errors += check_block(bl, hash_elements)
    if errors:
        errors.insert(0, f"\nErrors in screen {screen.name}, file name {module.__file__}:")
    return errors
    
def run_tests(user):    
    if config.autotest:
        for test_fn in Unishare.test_list:
            try:
                if asyncio.iscoroutinefunction(test_fn):
                    asyncio.run(test_fn())
                else: 
                    test_fn()
            except Exception as e:
                user.log(f'Test function {test_fn.__name__} failed with exception: {e}')
                
    errors = []
    for module in user.screens:
        module_errors = check_module(module)
        if module_errors: 
            errors += module_errors
        else:
            module.screen.set_reactivity(user) #because Users.py 95 does not set reactivity for screens for 0 user
    if errors:
        errors.insert(0, f'\n!!----Detected errors in screens:')
        print('\n'.join(errors), '\n')
    elif user.screens:
        print(f'\n----The screen definitions are correct.-----\n')
        
    files = config.autotest
    ok = True
    process = False
    if os.path.exists(testdir):
        for file in os.listdir(testdir):
            if not os.path.isdir(file) and (files == '*' or file in files):
                process = True
                if not test(file,user):
                    ok = False
    if process and ok:
        print('\n-----Autotests successfully passed.-----\n')  
    
    
        

            


# Copyright © 2024 UNISI Tech. All rights reserved.
from .units import Unit
from .common import *
from .dbunits import Dblist, dbupdates
from .llmrag import get_property
import asyncio
from collections import OrderedDict

relation_mark = 'Ⓡ'
exclude_mark = '✘'
max_len_rows4llm = 30

def get_chunk(obj, start_index):
    delta, data = obj.rows.get_delta_chunk(start_index)
    return {'update': 'updates', 'index': delta, 'data': data}

def accept_cell_value(table, dval: dict):            
    value = dval['value']
    if not isinstance(value, bool):
        try:
            value = float(value)        
        except:
            pass            
    if hasattr(table,'id'):
        dval['value'] = value
        if update := table.rows.update_cell(**dval):
            update['exclude'] = True       
    else:        
        table.rows[dval['delta']][dval['cell']] = value    
            
def delete_table_row(table, value):    
    if table.selected_list:
        if hasattr(table, 'link') and table.filter:
            link_table, rel_props, rel_name = table.rows.dbtable.list.link
            if not isinstance(value, list):                                
                value = [value]
            link_ids = [table.rows[index][-1] for index in value]
            table.rows.dbtable.delete_links(link_table.id, link_ids = link_ids, index_name = rel_name)
            table.__link_table_selection_changed__(link_table, link_table.value)
            return table
        elif isinstance(value, list):                    
            value.sort(reverse = True)
            for v in value:            
                del table.rows[v]
            table.value = []
        else:            
            del table.rows[value]  
            table.value = None    

def append_table_row(table, search_str = ''):
    ''' append has to return new row, value is the search string value in the table'''    
    new_row = [None] * len(table.headers)           
    if getattr(table,'id', None):          
        new_row = table.rows.append(new_row)        
        if hasattr(table, 'link') and table.filter:
            link_table, _, rel_name = table.rows.dbtable.list.link
            for linked_id in link_table.selected_list:
                relation = table.rows.dbtable.add_link(new_row[-1], link_table.id,
                     linked_id, link_index_name = rel_name) 
                new_row.extend(relation)                     
                break      
    else:           
        table.rows.append(new_row)
    return new_row

class Table(Unit):
    def __init__(self, *args, panda = None, **kwargs):
        if panda is not None:
            self._mark_changed = None
            self.mutate(PandaTable(*args, panda=panda, **kwargs))
        else:
            super().__init__(*args, **kwargs)    
            set_defaults(self, dict(headers = [], type = 'table', value = None, rows = [], editing = False, dense = True))
            self.__headers__ = self.headers[:]
        if hasattr(self,'id'):             
            if Unishare.db:
                Unishare.db.set_db_list(self)
            else:
                raise AssertionError('Config db_dir is not defined!')            
            self.get = get_chunk
            has_link = hasattr(self, 'link')
            set_defaults(self, {'filter': has_link, 'ids': False, 'search': ''})
            if has_link: 
                prop_types = {}
                rel_name = ''                
                match self.link:
                    case [link_table, prop_types, rel_name]: ...
                    case [link_table, prop_types]: ...
                    case link_table: ...
                rel_name, rel_fields = self.rows.dbtable.get_rel_fields2(link_table.id, prop_types, rel_name)
                if not hasattr(link_table, 'id'):
                    raise AttributeError('Linked table has to be persistent!')
                self.rows.link = link_table, list(prop_types.keys()), rel_name
                self.link = rel_fields
                
                @Unishare.handle(link_table,'changed')
                def link_table_selection_changed(master_table, val, init = False):
                    lstvalue = val if isinstance(val, list) else [val] if val != None else []
                    if lstvalue:
                        link_ids = [link_table.rows[val][-1] for val in lstvalue]
                        link_rows = self.rows.dbtable.calc_linked_rows(rel_name, link_ids, self.filter, self.search)
                    else:
                        link_rows = Dblist(self.rows.dbtable, cache = [])
                    if self.filter:                    
                        self.clean_selection()
                        self.rows = link_rows
                    else: 
                        selected_ids = [link_rows[i][-1] for i in range(len(link_rows))]
                        self.value = selected_ids  
                        if self.rows.cache is not None:
                            self.rows = self.rows.dbtable.list
                    if not init:
                        master_table.accept(val)              
                        return self     
                link_table_selection_changed(link_table, link_table.value, True)
                self.__link_table_selection_changed__ = link_table_selection_changed

                @Unishare.handle(self,'filter')
                def filter_status_changed(table, value):
                    self.filter = value
                    link_table_selection_changed(link_table, link_table.value, True)
                    self.calc_headers()
                    return self                
                
                @Unishare.handle(self,'changed')
                def changed_selection_causes__changing_links(self, new_value):                   
                    if link_table.value is not None and link_table.value != []:
                        #if link table is in multi mode, links are not editable 
                        if not self.filter and not isinstance(link_table.value, list | tuple):
                            if  self.editing:
                                actual = set(new_value if isinstance(new_value, list) else [] if new_value is None else [new_value])
                                old = set(self.value if isinstance(self.value, list) else ([] if self.value is None else [self.value]))                        
                                deselected = old - actual                        
                                if deselected:
                                    self.rows.dbtable.delete_links(link_table.id, link_table.value, deselected)                            
                                selected = actual - old
                                if selected:
                                    self.rows.dbtable.add_links(link_table.id, selected, link_table.value)                                                        
                            else:
                                return Warning('The linked table is not in edit mode', self)
                    return self.accept(new_value)    
            """
            @Unishare.handle(self,'search')
            def search_changed(table, value):
                self.search = value
                if has_link:
                    link_table_selection_changed(link_table, link_table.value, True)
                else:
                    self.rows = self.rows.dbtable.get_init_list(self.search)
                return self
            """        
            self.calc_headers()                                
                    
        elif hasattr(self,'ids'):
            raise ValueError("Only persistent tables can have 'ids' option!")

        if getattr(self,'edit', True): 
            set_defaults(self,{'delete': delete_table_row, 'append': append_table_row, 'modify': accept_cell_value})   

    @property
    def compact_view(self) -> str:
        """only selected are sended to llm"""
        selected = self.selected_list        
        if not selected and len(self.rows) < max_len_rows4llm:
            selected = range(len(self.rows))        
        str_rows = ';'.join(','.join(f'{field}: {value}' for field, value in zip(self.headers, self.rows[index])) for index in selected)
        return f'{self.name} : {str_rows}' 
    
    @property
    def selected_list(self):                            
        return [] if self.value is None else self.value if isinstance(self.value, list) else [self.value]   

    def clean_selection(self):        
        self.value = [] if isinstance(self.value,tuple | list) else None
        return self    
    
    @property
    def panda(self):
        if gp := getattr(self,'__panda__',None):
            return gp() 
    
    def calc_headers(self):        
        """only for persistent"""
        table_fields = self.rows.dbtable.table_fields
        self.headers = self.__headers__[:] if self.__headers__ else [pretty4(prop)for prop in table_fields]
        only_node_headers = len(self.headers) == len(table_fields)
        if self.ids:
            self.headers.insert(len(table_fields), 'ID')
        elif self.filter:
            self.headers.insert(len(table_fields), exclude_mark + 'ID')
        if self.filter:
            if only_node_headers:
                self.headers.extend([relation_mark + pretty4(link_field) for link_field in self.link])
            if self.ids:
                self.headers.append(relation_mark + 'ID')
    
    async def emit(self, *_):        
        """calcute llm field values for selected rows if they are None"""        
        if Unishare.llm_model and getattr(self, 'llm', None) is not None:              
            tasks = []
            for index in self.selected_list:
                values = {field: value for field, value in zip(self.headers, self.rows[index]) if value}
                for fld, deps in self._llm_dependencies.items():                    
                    if fld not in values:                        
                        if deps is True:
                            context = values
                        else:
                            context = OrderedDict()
                            for dep in deps:
                                value = values.get(dep, None)
                                if value is None:
                                    if self.llm: #exact
                                        continue   #not all fields
                                else:                                
                                    if isinstance(dep, str):
                                        context[dep] = value
                                    elif isinstance(dep, Unit):
                                        context[dep.name] = dep.value                                    
                                    else:
                                        raise AttributeError(f'Invalid llm parameter {dep} in {self.name} element!')
                        if context:                                                    
                            async def assign(index, fld, context):
                                self.rows[index][self.headers.index(fld)] = await get_property(fld, context)                            
                            context =  ','.join(f'{fld}: {val}' for fld, val in context.items())
                            tasks.append(asyncio.create_task(assign(index, fld, context)))
            if tasks:
                await asyncio.gather(*tasks)
                return self
    @property    
    def is_base_table_list(self):
        """is table in basic view mode"""
        if hasattr(self, 'id'):
            dbtable = self.rows.dbtable
            return dbtable.list is self.rows
        
def delete_panda_row(table, value):    
    pt = table.panda
    def delete_in_panda(row_index):
        if row_index < 0 or row_index >= len(pt):
            raise ValueError("Row number is out of range")
        pt.drop(index = row_index,  inplace=True)

    if isinstance(value, list | tuple):                    
        value.sort(reverse=True)
        for row_index in value:            
            delete_in_panda(row_index)        
    else:            
        delete_in_panda(value)        
    
    #pt.reset_index(inplace=True) 
    delete_table_row(table, value)    

def accept_panda_cell(table, value_pos: dict):
    value = value_pos['value']
    if not isinstance(value, bool):
        try:
            value = float(value)        
        except:
            pass                
    row_index, col_index = value_pos['delta'], value_pos['cell']
    table.panda.iat[row_index,col_index] = value
    accept_cell_value(table, value_pos)

def append_panda_row(table, row_index):    
    df = table.panda
    new_row = append_table_row(table, row_index)
    df.loc[len(df)] = new_row 
    #df.loc[len(df), df.columns] = new_row
    return new_row    

class PandaTable(Table):
    """ panda = opened panda table"""
    def __init__(self, name, *args, panda = None, fix_headers = True, **kwargs):
        Unit.__init__(self, name, *args, **kwargs)                  
        set_defaults(self, dict(type = 'table', value = None, editing = False, dense = True))        
        if panda is None:
            raise Exception('PandaTable has to get panda = pandaTable as an argument.')
        self.headers = panda.columns.tolist()
        if fix_headers:
            self.headers = [pretty4(header) for header in self.headers]        
        self.rows = panda.values.tolist()
        self.__panda__ = lambda: panda

        if getattr(self,'edit', True): 
            set_defaults(self,{'delete': delete_panda_row, 'append': append_panda_row,
                'modify': accept_panda_cell})
    
    